searchNodes=[{"doc":"Nebulex is split into 2 main components:Nebulex.Cache - caches are wrappers around the in-memory data store. Via the cache, we can put, get, update, delete and query existing entries. A cache needs an adapter to communicate to the in-memory data store.Nebulex.Caching - decorators provide an elegant way of annotating functions to be cached or evicted. By means of these decorators, it is possible the implementation of cache usage patterns like Read-through, Write-through, Cache-as-SoR, etc.In the following sections, we will provide an overview of those components and how they interact with each other. Feel free to access their respective module documentation for more specific examples, options and configuration.If you want to quickly check a sample application using Nebulex, please check the getting started guide.CachesNebulex.Cache is the wrapper around the Cache. We can define a cache as follows:defmodule MyApp.MyCache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local endWhere the configuration for the Cache must be in your application environment, usually defined in your config/config.exs:config :my_app, MyApp.MyCache, gc_interval: 3_600_000, #=&gt; 1 hr backend: :shards, partitions: 2Each cache in Nebulex defines a start_link/1 function that needs to be invoked before using the cache. In general, this function is not called directly, but used as part of your application supervision tree.If your application was generated with a supervisor (by passing --sup to mix new) you will have a lib/my_app/application.ex file containing the application start callback that defines and starts your supervisor. You just need to edit the start/2 function to start the cache as a supervisor on your application's supervisor:def start(_type, _args) do children = [ {MyApp.Cache, []} ] opts = [strategy: :one_for_one, name: MyApp.Supervisor] Supervisor.start_link(children, opts) endDeclarative annotation-based cachingSee Nebulex.Caching.","ref":"Nebulex.html","title":"Nebulex","type":"module"},{"doc":"This module specifies the adapter API that a Cache adapter is required to implement.","ref":"Nebulex.Adapter.html","title":"Nebulex.Adapter","type":"behaviour"},{"doc":"The callback invoked in case the adapter needs to inject code.","ref":"Nebulex.Adapter.html#c:__before_compile__/1","title":"Nebulex.Adapter.__before_compile__/1","type":"macrocallback"},{"doc":"Deletes a single entry from cache.See Nebulex.Cache.delete/2.","ref":"Nebulex.Adapter.html#c:delete/3","title":"Nebulex.Adapter.delete/3","type":"callback"},{"doc":"Returns true if the given key exists and the new ttl was successfully updated, otherwise, false is returned.See Nebulex.Cache.expire/2.","ref":"Nebulex.Adapter.html#c:expire/3","title":"Nebulex.Adapter.expire/3","type":"callback"},{"doc":"Flushes the cache and returns the number of evicted keys.See Nebulex.Cache.flush/0.","ref":"Nebulex.Adapter.html#c:flush/1","title":"Nebulex.Adapter.flush/1","type":"callback"},{"doc":"Gets the value for a specific key in cache.See Nebulex.Cache.get/2.","ref":"Nebulex.Adapter.html#c:get/3","title":"Nebulex.Adapter.get/3","type":"callback"},{"doc":"Gets a collection of entries from the Cache, returning them as Map.t() of the values associated with the set of keys requested.For every key that does not hold a value or does not exist, that key is simply ignored. Because of this, the operation never fails.See Nebulex.Cache.get_all/2.","ref":"Nebulex.Adapter.html#c:get_all/3","title":"Nebulex.Adapter.get_all/3","type":"callback"},{"doc":"Returns whether the given key exists in cache.See Nebulex.Cache.has_key?/1.","ref":"Nebulex.Adapter.html#c:has_key?/2","title":"Nebulex.Adapter.has_key?/2","type":"callback"},{"doc":"Increments or decrements the counter mapped to the given key.See Nebulex.Cache.incr/3.","ref":"Nebulex.Adapter.html#c:incr/5","title":"Nebulex.Adapter.incr/5","type":"callback"},{"doc":"Initializes the adapter supervision tree by returning the children.","ref":"Nebulex.Adapter.html#c:init/1","title":"Nebulex.Adapter.init/1","type":"callback"},{"doc":"Puts the given value under key into the cache.Returns true if the value with key key is successfully inserted; otherwise false is returned.The ttl argument sets the time-to-live for the stored entry. If it is not set, it means the entry hasn't a time-to-live, then it shouldn't expire.OnWriteThe on_write argument supports the following values::put - If the key already exists, it is overwritten. Any previous time-to-live associated with the key is discarded on successful write operation.:put_new - It only stores the entry if the key does not already exist, otherwise, false is returned.:replace - Alters the value stored under the given key, but only if the key already exists into the cache, otherwise, false is returned.See Nebulex.Cache.put/3, Nebulex.Cache.put_new/3, Nebulex.Cache.replace/3.","ref":"Nebulex.Adapter.html#c:put/6","title":"Nebulex.Adapter.put/6","type":"callback"},{"doc":"Puts the given entries (key/value pairs) into the cache.Returns true if all the keys were inserted. If no key was inserted (at least one key already existed), false is returned.The ttl argument sets the time-to-live for the stored entry. If it is not set, it means the entry hasn't a time-to-live, then it shouldn't expire. The given ttl is applied to all keys.OnWriteThe on_write argument supports the following values::put - If the key already exists, it is overwritten. Any previous time-to-live associated with the key is discarded on successful write operation.:put_new - It only stores the entry if the key does not already exist, otherwise, false is returned.Ideally, this operation should be atomic, so all given keys are set at once. But it depends purely on the adapter's implementation and the backend used internally by the adapter. Hence, it is recommended to checkout the adapter's documentation.See Nebulex.Cache.put_all/2.","ref":"Nebulex.Adapter.html#c:put_all/5","title":"Nebulex.Adapter.put_all/5","type":"callback"},{"doc":"Returns the total number of cached entries.See Nebulex.Cache.size/0.","ref":"Nebulex.Adapter.html#c:size/1","title":"Nebulex.Adapter.size/1","type":"callback"},{"doc":"Returns and removes the entry with key key in the cache.See Nebulex.Cache.take/2.","ref":"Nebulex.Adapter.html#c:take/3","title":"Nebulex.Adapter.take/3","type":"callback"},{"doc":"Returns true if the given key exists and the last access time was successfully updated, otherwise, false is returned.See Nebulex.Cache.touch/1.","ref":"Nebulex.Adapter.html#c:touch/2","title":"Nebulex.Adapter.touch/2","type":"callback"},{"doc":"Returns the TTL (time-to-live) for the given key. If the key does not exist, then nil is returned.See Nebulex.Cache.ttl/1.","ref":"Nebulex.Adapter.html#c:ttl/2","title":"Nebulex.Adapter.ttl/2","type":"callback"},{"doc":"RExecutes the function fun passing as parameters the adapter and metadata (from the init/1 callback) associated with the given cache name_or_pid.It expects a name or a PID representing the cache.","ref":"Nebulex.Adapter.html#with_meta/2","title":"Nebulex.Adapter.with_meta/2","type":"function"},{"doc":"The metadata returned by the adapter init/1.It must be a map and Nebulex itself will always inject two keys into the meta::cache - The cache module.:name - The nase of the cache.:pid - The PID returned by the child spec returned in init/1","ref":"Nebulex.Adapter.html#t:adapter_meta/0","title":"Nebulex.Adapter.adapter_meta/0","type":"type"},{"doc":"","ref":"Nebulex.Adapter.html#t:cache/0","title":"Nebulex.Adapter.cache/0","type":"type"},{"doc":"","ref":"Nebulex.Adapter.html#t:child_spec/0","title":"Nebulex.Adapter.child_spec/0","type":"type"},{"doc":"","ref":"Nebulex.Adapter.html#t:entries/0","title":"Nebulex.Adapter.entries/0","type":"type"},{"doc":"","ref":"Nebulex.Adapter.html#t:entry/0","title":"Nebulex.Adapter.entry/0","type":"type"},{"doc":"","ref":"Nebulex.Adapter.html#t:key/0","title":"Nebulex.Adapter.key/0","type":"type"},{"doc":"","ref":"Nebulex.Adapter.html#t:on_write/0","title":"Nebulex.Adapter.on_write/0","type":"type"},{"doc":"","ref":"Nebulex.Adapter.html#t:opts/0","title":"Nebulex.Adapter.opts/0","type":"type"},{"doc":"","ref":"Nebulex.Adapter.html#t:t/0","title":"Nebulex.Adapter.t/0","type":"type"},{"doc":"","ref":"Nebulex.Adapter.html#t:ttl/0","title":"Nebulex.Adapter.ttl/0","type":"type"},{"doc":"","ref":"Nebulex.Adapter.html#t:value/0","title":"Nebulex.Adapter.value/0","type":"type"},{"doc":"This behaviour provides a callback to compute the hash slot for a specific key based on the number of slots/nodes.The purpose of this module is to allow users to implement a custom hash function to distribute keys. It can be used to select the node/slot where a specific key is supposed to be.It is highly recommended to use a Consistent Hashing algorithm.Exampledefmodule MyApp.HashSlot do @behaviour Nebulex.Adapter.HashSlot def keyslot(key, range) do key |&gt; :erlang.phash2() |&gt; :jchash.compute(range) end endThis example uses Jumping Consistent Hash.","ref":"Nebulex.Adapter.HashSlot.html","title":"Nebulex.Adapter.HashSlot","type":"behaviour"},{"doc":"Computes the hash for a specific key based on the number of slots given by range. Returns a slot within the range 0..range-1.Exampleiex&gt; MyHash.keyslot(&quot;mykey&quot;, 10) 2","ref":"Nebulex.Adapter.HashSlot.html#c:keyslot/2","title":"Nebulex.Adapter.HashSlot.keyslot/2","type":"callback"},{"doc":"Specifies the adapter persistence API.Default implementationThis module provides a default implementation that uses File and Stream under-the-hood. For dumping a cache to a file, the entries are streamed from the cache and written in chunks (one chunk per line), and each chunk contains N number of entries. For loading the entries from a file, the file is read and streamed line-by-line, so that the entries collected on each line are inserted in streaming fashion as well.The default implementation accepts the following options only for dump operation (there are not options for load):entries_per_line - The number of entries to be written per line in the file. Defaults to 10.compression - The compression level. The values are the same as :erlang.term_to_binary /2. Defaults to 6.See Nebulex.Cache.dump/2 and Nebulex.Cache.load/2 for more information.","ref":"Nebulex.Adapter.Persistence.html","title":"Nebulex.Adapter.Persistence","type":"behaviour"},{"doc":"Dumps a cache to the given file path.Returns :ok if successful, or {:error, reason} if an error occurs.See Nebulex.Cache.dump/2.","ref":"Nebulex.Adapter.Persistence.html#c:dump/3","title":"Nebulex.Adapter.Persistence.dump/3","type":"callback"},{"doc":"Loads a dumped cache from the given path.Returns :ok if successful, or {:error, reason} if an error occurs.See Nebulex.Cache.load/2.","ref":"Nebulex.Adapter.Persistence.html#c:load/3","title":"Nebulex.Adapter.Persistence.load/3","type":"callback"},{"doc":"Specifies the query API required from adapters.","ref":"Nebulex.Adapter.Queryable.html","title":"Nebulex.Adapter.Queryable","type":"behaviour"},{"doc":"Fetches all entries from cache matching the given query.If the query is nil, it fetches all entries from cache; this is common for all adapters. However, the query could be any other value, that depends entirely on the adapter's implementation. Therefore, it is recommended to check out adapters' documentation. For instance, the built-in Nebulex.Adapters.Local adapter supports :ets.match_spec() as query.May raise Nebulex.QueryError if query validation fails.See Nebulex.Cache.all/2.","ref":"Nebulex.Adapter.Queryable.html#c:all/3","title":"Nebulex.Adapter.Queryable.all/3","type":"callback"},{"doc":"Streams the given query.It returns a stream of values.May raise Nebulex.QueryError if query validation fails.See Nebulex.Cache.stream/2.","ref":"Nebulex.Adapter.Queryable.html#c:stream/3","title":"Nebulex.Adapter.Queryable.stream/3","type":"callback"},{"doc":"Specifies the adapter transactions API.Default implementationThis module also provides a default implementation which uses the Erlang library :global.This implementation accepts the following options::keys - The list of the keys that will be locked. Since the lock id is generated based on the key, if this option is not set, a fixed/constant lock id is used to perform the transaction, then all further transactions (without this option set) are serialized and the performance is affected significantly. For that reason it is recommended to pass the list of keys involved in the transaction.:nodes - The list of the nodes where the lock will be set, or on all nodes if none are specified.:retries - If the key has been locked by other process already, and :retries is not equal to 0, the process sleeps for a while and tries to execute the action later. When :retries attempts have been made, an exception is raised. If :retries is :infinity (the default), the function will eventually be executed (unless the lock is never released).Let's see an example:MyCache.transaction fn -&gt; counter = MyCache.get(:counter) MyCache.set(:counter, counter + 1) endLocking only the involved key (recommended):MyCache.transaction [keys: [:counter]], fn -&gt; counter = MyCache.get(:counter) MyCache.set(:counter, counter + 1) end MyCache.transaction [keys: [:alice, :bob]], fn -&gt; alice = MyCache.get(:alice) bob = MyCache.get(:bob) MyCache.set(:alice, %{alice | balance: alice.balance + 100}) MyCache.set(:bob, %{bob | balance: bob.balance + 100}) end","ref":"Nebulex.Adapter.Transaction.html","title":"Nebulex.Adapter.Transaction","type":"behaviour"},{"doc":"Returns true if the given process is inside a transaction.See Nebulex.Cache.in_transaction?/0.","ref":"Nebulex.Adapter.Transaction.html#c:in_transaction?/1","title":"Nebulex.Adapter.Transaction.in_transaction?/1","type":"callback"},{"doc":"Runs the given function inside a transaction.A successful transaction returns the value returned by the function.See Nebulex.Cache.transaction/2.","ref":"Nebulex.Adapter.Transaction.html#c:transaction/3","title":"Nebulex.Adapter.Transaction.transaction/3","type":"callback"},{"doc":"Adapter module for Local Generational Cache; inspired by epocxy.Generational caching using an ets table (or multiple ones when used with :shards) for each generation of cached data. Accesses hit the newer generation first, and migrate from the older generation to the newer generation when retrieved from the stale table. When a new generation is started, the oldest one is deleted. This is a form of mass garbage collection which avoids using timers and expiration of individual cached elements.This implementation of generation cache uses only two generations (which is more than enough) also referred like the newer and the older.FeaturesConfigurable backend (ets or :shards).Expiration – A status based on TTL (Time To Live) option. To maintain cache performance, expired entries may not be immediately flushed or evicted, they are expired or evicted on-demand, when the key is read.Eviction – Generational Garbage Collection.Sharding – For intensive workloads, the Cache may also be partitioned (by using :shards backend and specifying the :partitions option).Support for transactions via Erlang global name registration facility.OptionsThis adapter supports the following options and all of them can be given via the cache configuration::backend - Defines the backend or storage to be used for the adapter. Supported backends are: :ets and :shards. Defaults to :ets.:read_concurrency - Since this adapter uses ETS tables internally, this option is used when a new table is created. See :ets.new/2. Defaults to true.:write_concurrency - Since this adapter uses ETS tables internally, this option is used when a new table is created. See :ets.new/2. Defaults to true.:compressed - This option is used when a new ETS table is created and it defines whether or not it includes X as an option. See :ets.new/2. Defaults to false.:backend_type - This option defines the type of ETS to be used (Defaults to :set). However, it is highly recommended to keep the default value, since there are commands not supported (unexpected exception may be raised) for types like :bag or : duplicate_bag. Please see the ETS docs for more information.:partitions - The number of partitions in the Cache. This option is only available for :shards backend. Defaults to System.schedulers_online().:gc_interval - Interval time in milliseconds to garbage collection to run, delete the oldest generation and create a new one. If this option is not set, garbage collection is never executed, so new generations must be created explicitly, e.g.: Generation.new_generation(cache_name, []).:max_size - Max number of cached entries (cache limit). If it is not set (nil), the check to release memory is not performed (the default).:allocated_memory - Max size in bytes allocated for a cache generation. If this option is set and the configured value is reached, a new cache generation is created so the oldest is deleted and force releasing memory space. If it is not set (nil), the cleanup check to release memory is not performed (the default).:gc_cleanup_min_timeout - The min timeout in milliseconds for triggering the next cleanup and memory check. This will be the timeout to use when the max allocated memory is reached. Defaults to 30_000.:gc_cleanup_max_timeout - The max timeout in milliseconds for triggering the next cleanup and memory check. This is the timeout used when the cache starts or the consumed memory is 0. Defaults to 300_000.ExampleNebulex.Cache is the wrapper around the cache. We can define a local cache as follows:defmodule MyApp.LocalCache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local endWhere the configuration for the cache must be in your application environment, usually defined in your config/config.exs:config :my_app, MyApp.LocalCache, backend: :shards, gc_interval: 86_400_000, max_size: 200_000, allocated_memory: 2_000_000_000, gc_cleanup_min_timeout: 10_000, gc_cleanup_max_timeout: 600_000For intensive workloads, the Cache may also be partitioned (by using :shards backend and specifying the :partitions option). If partitioning is required then a good default is to set the number of partitions to the number of schedulers available (the default):config :my_app, MyApp.LocalCache, backend: :shards, gc_interval: 86_400_000, max_size: 200_000, allocated_memory: 2_000_000_000, gc_cleanup_min_timeout: 10_000, gc_cleanup_max_timeout: 600_000, partitions: System.schedulers_online()For more information about the usage, check out Nebulex.Cache.Queryable APIThe adapter supports as query parameter the following values:query - nil | :unexpired | :expired | :ets.match_spec()Internally, an entry is represented by the tuple {key, val, vsn, exp}, which means the match pattern within the :ets.match_spec() must be something like {:&quot;$1&quot;, :&quot;$2&quot;, :&quot;$3&quot;, :&quot;$4&quot;}. In order to make query building easier, you can use Ex2ms library.Examples# built-in queries MyCache.all() MyCache.all(:unexpired) MyCache.all(:expired) # using a custom match spec (all values &gt; 10) spec = [{{:&quot;$1&quot;, :&quot;$2&quot;, :_, :_}, [{:&gt;, :&quot;$2&quot;, 10}], [{{:&quot;$1&quot;, :&quot;$2&quot;}}]}] MyCache.all(spec) # using Ex2ms import Ex2ms spec = fun do {key, value, _version, _expire_at} when value &gt; 10 -&gt; {key, value} end MyCache.all(spec)The :return option applies only for built-in queries, such as: nil | :unexpired | :expired, if you are using a custom :ets.match_spec(), the return value depends on it.The same applies to the stream function.","ref":"Nebulex.Adapters.Local.html","title":"Nebulex.Adapters.Local","type":"module"},{"doc":"","ref":"Nebulex.Adapters.Local.html#entry/1","title":"Nebulex.Adapters.Local.entry/1","type":"macro"},{"doc":"","ref":"Nebulex.Adapters.Local.html#entry/2","title":"Nebulex.Adapters.Local.entry/2","type":"macro"},{"doc":"Generational garbage collection process.The generational garbage collector manage the heap as several sub-heaps, known as generations, based on age of the objects. An object is allocated in the youngest generation, sometimes called the nursery, and is promoted to an older generation if its lifetime exceeds the threshold of its current generation (defined by option :gc_interval). Everytime the GC runs (triggered by :gc_interval timeout), a new cache generation is created and the oldest one is deleted.The only way to create new generations is through this module (this server is the metadata owner) calling new/2 function. When a Cache is created, a generational garbage collector is attached to it automatically, therefore, this server MUST NOT be started directly.OptionsThese options are configured through the Nebulex.Adapters.Local adapter::gc_interval - Interval time in milliseconds to garbage collection to run, delete the oldest generation and create a new one. If this option is not set, garbage collection is never executed, so new generations must be created explicitly, e.g.: new(cache, []).:max_size - Max number of cached entries (cache limit). If it is not set (nil), the check to release memory is not performed (the default).:allocated_memory - Max size in bytes allocated for a cache generation. If this option is set and the configured value is reached, a new cache generation is created so the oldest is deleted and force releasing memory space. If it is not set (nil), the cleanup check to release memory is not performed (the default).:gc_cleanup_min_timeout - The min timeout in milliseconds for triggering the next cleanup and memory check. This will be the timeout to use when the max allocated memory is reached. Defaults to 30_000.:gc_cleanup_max_timeout - The max timeout in milliseconds for triggering the next cleanup and memory check. This is the timeout used when the cache starts or the consumed memory is 0. Defaults to 300_000.","ref":"Nebulex.Adapters.Local.Generation.html","title":"Nebulex.Adapters.Local.Generation","type":"module"},{"doc":"Returns a specification to start this module under a supervisor.See Supervisor.","ref":"Nebulex.Adapters.Local.Generation.html#child_spec/1","title":"Nebulex.Adapters.Local.Generation.child_spec/1","type":"function"},{"doc":"Flushes the cache (including all its generations).ExampleNebulex.Adapters.Local.Generation.flush(MyCache)","ref":"Nebulex.Adapters.Local.Generation.html#flush/1","title":"Nebulex.Adapters.Local.Generation.flush/1","type":"function"},{"doc":"Returns the list of the generations in the form [newer, older].ExampleNebulex.Adapters.Local.Generation.list(MyCache)","ref":"Nebulex.Adapters.Local.Generation.html#list/1","title":"Nebulex.Adapters.Local.Generation.list/1","type":"function"},{"doc":"Returns the memory info in a tuple {used_mem, total_mem}.ExampleNebulex.Adapters.Local.Generation.memory_info(MyCache)","ref":"Nebulex.Adapters.Local.Generation.html#memory_info/1","title":"Nebulex.Adapters.Local.Generation.memory_info/1","type":"function"},{"doc":"Creates a new cache generation. Once the max number of generations is reached, when a new generation is created, the oldest one is deleted.Options:reset_timer - Indicates if the poll frequency time-out should be reset or not (default: true).ExampleNebulex.Adapters.Local.Generation.new(MyCache, reset_timer: :false)","ref":"Nebulex.Adapters.Local.Generation.html#new/2","title":"Nebulex.Adapters.Local.Generation.new/2","type":"function"},{"doc":"Returns the newer generation.ExampleNebulex.Adapters.Local.Generation.newer(MyCache)","ref":"Nebulex.Adapters.Local.Generation.html#newer/1","title":"Nebulex.Adapters.Local.Generation.newer/1","type":"function"},{"doc":"Reallocates the block of memory that was previously allocated for the given cache with the new size. In other words, reallocates the max memory size for a cache generation.ExampleNebulex.Adapters.Local.Generation.realloc(MyCache, 1_000_000)","ref":"Nebulex.Adapters.Local.Generation.html#realloc/2","title":"Nebulex.Adapters.Local.Generation.realloc/2","type":"function"},{"doc":"Returns the name of the GC server for the given cache name.ExampleNebulex.Adapters.Local.Generation.server_name(MyCache)","ref":"Nebulex.Adapters.Local.Generation.html#server_name/1","title":"Nebulex.Adapters.Local.Generation.server_name/1","type":"function"},{"doc":"Starts the garbage collector for the build-in local cache adapter.","ref":"Nebulex.Adapters.Local.Generation.html#start_link/1","title":"Nebulex.Adapters.Local.Generation.start_link/1","type":"function"},{"doc":"Adapter module for Multi-level Cache.This is just a simple layer on top of local or distributed cache implementations that enables to have a cache hierarchy by levels. Multi-level caches generally operate by checking the fastest, level 1 (L1) cache first; if it hits, the adapter proceeds at high speed. If that first cache misses, the next fastest cache (level 2, L2) is checked, and so on, before accessing external memory (that can be handled by a :fallback function).For write functions, the &quot;Write Through&quot; policy is applied by default; this policy ensures that the data is stored safely as it is written throughout the hierarchy. However, it is possible to force the write operation in a specific level (although it is not recommended) via level option, where the value is a positive integer greater than 0.We can define a multi-level cache as follows:defmodule MyApp.Multilevel do use Nebulex.Cache, otp_app: :nebulex, adapter: Nebulex.Adapters.Multilevel endWhere the configuration for the Cache must be in your application environment, usually defined in your config/config.exs:config :my_app, MyApp.Multilevel, model: :inclusive, levels: [ l1: [ gc_interval: 86_400_000, backend: :shards, partitions: 2 ], l2: [ adapter: Nebulex.Adapters.Partitioned, primary: [ gc_interval: 86_400_000, backend: :shards, partitions: 2 ] ] ]For more information about the usage, see Nebulex.Cache documentation.OptionsThis adapter supports the following options and all of them can be given via the cache configuration::levels - A keyword list that defines each cache level configuration. The key is used as the name for that cache level, and the value is the level config, which depends on the adapter to use. The order in which the levels are defined is the same the multi-level cache will use. For example, the first cache in the list will be the L1 cache (level 1) and so on; the Nth elemnt will be the LN cache. This option is mandatory, if it is not set or empty, an exception will be raised. See &quot;Shared Level Options&quot; below.:model - Specifies the cache model: :inclusive or :exclusive; defaults to :inclusive. In an inclusive cache, the same data can be present in all caches/levels. In an exclusive cache, data can be present in only one cache/level and a key cannot be found in the rest of caches at the same time. This option affects get operation only; if :cache_model is :inclusive, when the key is found in a level N, that entry is duplicated backwards (to all previous levels: 1..N-1).Shared Level Options:adapter - The adapter to be used for the configured level. Defaults to Nebulex.Adapters.Local.The rest of the options depend on the adapter to use.Run-Time OptionsSome of the cache functions accept the following runtime options::level - It may be an integer greater than 0 that specifies the cache level where the operation will take place. By default, the evaluation is performed throughout the whole cache hierarchy (all levels).Extended APIThis adapter provides one additional convenience function for retrieving the cache model for the given cache name:MyCache.model() MyCache.model(:cache_name)LimitationsBecause this adapter reuses other existing/configured adapters, it inherits all their limitations too. Therefore, it is highly recommended to check the documentation of the adapters to use.","ref":"Nebulex.Adapters.Multilevel.html","title":"Nebulex.Adapters.Multilevel","type":"module"},{"doc":"Built-in adapter for partitioned cache topology.A partitioned cache is a clustered, fault-tolerant cache that has linear scalability. Data is partitioned among all the machines of the cluster. For fault-tolerance, partitioned caches can be configured to keep each piece of data on one or more unique machines within a cluster. This adapter in particular hasn't fault-tolerance built-in, each piece of data is kept in a single node/machine (sharding), therefore, if a node fails, the data kept by this node won't be available for the rest of the cluster.PG2 is used under-the-hood by the adapter to manage the cluster nodes. When the partitioned cache is started in a node, it creates a PG2 group and joins it (the cache supervisor PID is joined to the group). Then, when a function is invoked, the adapter picks a node from the node list (using the PG2 group members), and then the function is executed on that node. In the same way, when the supervisor process of the partitioned cache dies, the PID of that process is automatically removed from the PG2 group; this is why it's recommended to use a consistent hashing algorithm for the node selector.NOTE: pg2 will be replaced by pg in future, since the pg2 module is deprecated as of OTP 23 and scheduled for removal in OTP 24.This adapter depends on a local cache adapter (primary storage), it adds a thin layer on top of it in order to distribute requests across a group of nodes, where is supposed the local cache is running already. However, you don't need to define or declare an additional cache module for the ocal store, instead, the adapter initializes it automatically (adds the local cache store as part of the supervision tree) based on the given options within the primary: argument.FeaturesSupport for partitioned topology (Sharding Distribution Model).Support for transactions via Erlang global name registration facility.Configurable primary store (primary local cache).Configurable hash-slot module to compute the node.We can define a partitioned cache as follows:defmodule MyApp.PartitionedCache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Partitioned @behaviour Nebulex.Adapter.HashSlot @impl true def keyslot(key, range) do key |&gt; :erlang.phash2() |&gt; :jchash.compute(range) end endWhere the configuration for the cache must be in your application environment, usually defined in your config/config.exs:config :my_app, MyApp.PartitionedCache, hash_slot: MyApp.PartitionedCache primary: [ adapter: Nebulex.Adapters.Local, gc_interval: 86_400_000, backend: :shards, partitions: System.schedulers_online() ]For more information about the usage, see Nebulex.Cache documentation.OptionsThis adapter supports the following options and all of them can be given via the cache configuration::primary - The options that will be passed to the adapter associated with the local primary store. These options depend on the adapter to use, except for the shared option adapter: (see shared primary options below).:hash_slot - Defines the module implementing Nebulex.Adapter.HashSlot behaviour.task_supervisor_opts - Start-time options passed to Task.Supervisor.start_link/1 when the adapter is initialized.Shared Primary Options:adapter - The adapter to be used for the partitioned cache as the local primary store. Defaults to Nebulex.Adapters.Local.The rest of the options depend on the adapter to use.Runtime optionsThese options apply to all adapter's functions.:timeout - The time-out value in milliseconds for the command that will be executed. If the timeout is exceeded, then the current process will exit. For executing a command on remote nodes, this adapter uses Task.await/2 internally for receiving the result, so this option tells how much time the adapter should wait for it. If the timeout is exceeded, the task is shut down but the current process doesn't exit, only the result associated with that task is skipped in the reduce phase.Extended APIThis adapter provides some additional convenience functions to the Nebulex.Cache API.Retrieving the cluster nodes associated with the given cache name:MyCache.nodes() MyCache.nodes(:cache_name)Get a cluster node for the cache name based on the given key:MyCache.get_node(&quot;mykey&quot;) MyCache.get_node(:cache_name, &quot;mykey&quot;)If no cache name is passed to the previous functions, the name of the calling cache module is used by defaultLimitationsFor Nebulex.Cache.get_and_update/3 and Nebulex.Cache.update/4, they both have a parameter that is the anonymous function, and it is compiled into the module where it is created, which means it necessarily doesn't exists on remote nodes. To ensure they work as expected, you must provide functions from modules existing in all nodes of the group.","ref":"Nebulex.Adapters.Partitioned.html","title":"Nebulex.Adapters.Partitioned","type":"module"},{"doc":"Helper to perform stream/3 locally.","ref":"Nebulex.Adapters.Partitioned.html#eval_local_stream/4","title":"Nebulex.Adapters.Partitioned.eval_local_stream/4","type":"function"},{"doc":"","ref":"Nebulex.Adapters.Partitioned.html#get_node/2","title":"Nebulex.Adapters.Partitioned.get_node/2","type":"function"},{"doc":"Built-in adapter for replicated cache topology.The replicated cache excels in its ability to handle data replication, concurrency control and failover in a cluster, all while delivering in-memory data access speeds. A clustered replicated cache is exactly what it says it is: a cache that replicates its data to all cluster nodes.There are several challenges to building a reliably replicated cache. The first is how to get it to scale and perform well. Updates to the cache have to be sent to all cluster nodes, and all cluster nodes have to end up with the same data, even if multiple updates to the same piece of data occur at the same time. Also, if a cluster node requests a lock, ideally it should not have to get all cluster nodes to agree on the lock or at least do it in a very efficient way (:global is used for this), otherwise it will scale extremely poorly; yet in the case of a cluster node failure, all of the data and lock information must be kept safely.The best part of a replicated cache is its access speed. Since the data is replicated to each cluster node, it is available for use without any waiting. This is referred to as &quot;zero latency access,&quot; and is perfect for situations in which an application requires the highest possible speed in its data access.However, there are some limitations:&lt;ins&gt;Cost Per Update&lt;/ins&gt; - Updating a replicated cache requires pushing the new version of the data to all other cluster members, which will limit scalability if there is a high frequency of updates per member.&lt;ins&gt;Cost Per Entry&lt;/ins&gt; - The data is replicated to every cluster member, so Memory Heap space is used on each member, which will impact performance for large caches.Based on &quot;Distributed Caching Essential Lessons&quot; by Cameron Purdy.We can define a replicated cache as follows:defmodule MyApp.ReplicatedCache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Replicated endWhere the configuration for the cache must be in your application environment, usually defined in your config/config.exs:config :my_app, MyApp.ReplicatedCache, primary: [ adapter: Nebulex.Adapters.Local, gc_interval: 86_400_000, backend: :shards, partitions: System.schedulers_online() ]For more information about the usage, see Nebulex.Cache documentation.OptionsThis adapter supports the following options and all of them can be given via the cache configuration::primary - The options that will be passed to the adapter associated with the local primary store. These options depend on the adapter to use, except for the shared option adapter: (see shared primary options below).task_supervisor_opts - Start-time options passed to Task.Supervisor.start_link/1 when the adapter is initialized.:bootstrap_timeout - a timeout in milliseconds that bootstrap process will wait after the cache supervision tree is started so the data can be imported from remote nodes. Defaults to 1000.Shared Primary Options:adapter - The adapter to be used for the replicated cache as the local primary store. Defaults to Nebulex.Adapters.Local.The rest of the options depend on the adapter to use.Runtime optionsThese options apply to all adapter's functions.:timeout - The time-out value in milliseconds for the command that will be executed. If the timeout is exceeded, then the current process will exit. For executing a command on remote nodes, this adapter uses Task.await/2 internally for receiving the result, so this option tells how much time the adapter should wait for it. If the timeout is exceeded, the task is shut down but the current process doesn't exit, only the result associated with that task is skipped in the reduce phase.Extended APIThis adapter provides one additional convenience function for retrieving the cluster nodes associated with the given cache name:MyCache.nodes() MyCache.nodes(:cache_name)","ref":"Nebulex.Adapters.Replicated.html","title":"Nebulex.Adapters.Replicated","type":"module"},{"doc":"Cache's main interface; defines the cache abstraction layer which is highly inspired by Ecto.A Cache maps to an underlying implementation, controlled by the adapter. For example, Nebulex ships with a default adapter that implements a local generational cache.When used, the Cache expects the :otp_app and :adapter as options. The :otp_app should point to an OTP application that has the Cache configuration. For example, the Cache:defmodule MyCache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local endCould be configured with:config :my_app, MyCache, stats: true, backend: :shards, gc_interval: 86_400_000, #=&gt; 1 day max_size: 200_000, gc_cleanup_min_timeout: 10_000, gc_cleanup_max_timeout: 900_000, partitions: System.schedulers_online()Most of the configuration that goes into the config is specific to the adapter. For this particular example, you can check Nebulex.Adapters.Local for more information. In spite of this, the following configuration values are shared across all adapters::name- The name of the Cache supervisor process (Optional). If it is not passed within the options, the name of the cache module will be used as the name by default.:stats - The stats are supposed to be handled by the adapters, hence, it is recommended to check the adapters' documentation for supported stats, config, and so on. Nevertheless, Nebulex built-in adapters provide support for stats by setting the :stats option to true (Defaults to false). You can get the stats info by calling Nebulex.Cache.Stats.info(cache_or_name) at any time. For more information, See Nebulex.Cache.Stats.NOTE: It is highly recommendable to check the adapters' documentation.","ref":"Nebulex.Cache.html","title":"Nebulex.Cache","type":"behaviour"},{"doc":"Returns the adapter tied to the cache.","ref":"Nebulex.Cache.html#c:__adapter__/0","title":"Nebulex.Cache.__adapter__/0","type":"callback"},{"doc":"Fetches all entries from cache matching the given query.If the query is nil, it fetches all entries from cache; this is common for all adapters. However, the query could be any other value, which depends entirely on the adapter's implementation; see the &quot;Query&quot; section below.May raise Nebulex.QueryError if query validation fails.Options:return - Tells the query what to return from the matched entries. The possible values are: :key, :value, and :entry ({key, value} pairs). Defaults to :key. This option is supported by the build-in adapters, but it is recommended to check the adapter's documentation to confirm its compatibility with this option.See the &quot;Shared options&quot; section at the module documentation for more options.Example# set some entries iex&gt; :ok = Enum.each(1..5, &amp;MyCache.put(&amp;1, &amp;1 * 2)) # fetch all (with default params) iex&gt; MyCache.all() [1, 2, 3, 4, 5] # fetch all entries and return values iex&gt; MyCache.all(nil, return: :value) [2, 4, 6, 8, 10] # fetch all entries that match with the given query # assuming we are using Nebulex.Adapters.Local adapter iex&gt; query = [{{:_, :&quot;$1&quot;, :&quot;$2&quot;, :_, :_}, [{:&gt;, :&quot;$2&quot;, 5}], [:&quot;$1&quot;]}] iex&gt; MyCache.all(query) [3, 4, 5]QueryQuery spec is defined by the adapter, hence, it is recommended to review adapters documentation. For instance, the built-in Nebulex.Adapters.Local adapter supports nil | :unexpired | :expired | :ets.match_spec() as query value.Examples# additional built-in queries for Nebulex.Adapters.Local adapter iex&gt; unexpired = MyCache.all(:unexpired) iex&gt; expired = MyCache.all(:expired) # if we are using Nebulex.Adapters.Local adapter, the stored entry tuple # {:entry, key, value, version, expire_at}, then the match spec could be # something like: iex&gt; spec = [{{:entry, :&quot;$1&quot;, :&quot;$2&quot;, :_, :_}, [{:&gt;, :&quot;$2&quot;, 5}], [{{:&quot;$1&quot;, :&quot;$2&quot;}}]}] iex&gt; MyCache.all(spec) [{3, 6}, {4, 8}, {5, 10}] # the same previous query but using Ex2ms iex&gt; import Ex2ms Ex2ms iex&gt; spec = ...&gt; fun do ...&gt; {_. key, value, _, _} when value &gt; 5 -&gt; {key, value} ...&gt; end iex&gt; MyCache.all(spec) [{3, 6}, {4, 8}, {5, 10}]","ref":"Nebulex.Cache.html#c:all/2","title":"Nebulex.Cache.all/2","type":"callback"},{"doc":"Returns the adapter configuration stored in the :otp_app environment.If the c:init/2 callback is implemented in the cache, it will be invoked.","ref":"Nebulex.Cache.html#c:config/0","title":"Nebulex.Cache.config/0","type":"callback"},{"doc":"Deletes the entry in Cache for a specific key.OptionsSee the &quot;Shared options&quot; section at the module documentation for more options.Exampleiex&gt; MyCache.put(:a, 1) :ok iex&gt; MyCache.delete(:a) :ok iex&gt; MyCache.get(:a) :ok iex&gt; MyCache.delete(:non_existent_key) :ok","ref":"Nebulex.Cache.html#c:delete/2","title":"Nebulex.Cache.delete/2","type":"callback"},{"doc":"Dumps a cache to the given file path.Returns :ok if successful, or {:error, reason} if an error occurs.OptionsThis operation relies entirely on the adapter implementation, which means the options depend on each of them. For that reason, it is recommended to review the documentation of the adapter to be used. The built-in adapters inherit the default implementation from Nebulex.Adapter.Persistence, hence, review the available options there.Examples# set some entries iex&gt; entries = for x &lt;- 1..10, into: %{}, do: {x, x} iex&gt; MyCache.set_many(entries) :ok # dump cache to a file iex&gt; MyCache.dump(&quot;my_cache&quot;) :ok","ref":"Nebulex.Cache.html#c:dump/2","title":"Nebulex.Cache.dump/2","type":"callback"},{"doc":"Returns true if the given key exists and the new ttl was successfully updated, otherwise, false is returned.Examplesiex&gt; MyCache.put(:a, 1) :ok iex&gt; MyCache.expire(:a, 5) true iex&gt; MyCache.expire(:a, :infinity) true iex&gt; MyCache.ttl(:b, 5) false","ref":"Nebulex.Cache.html#c:expire/2","title":"Nebulex.Cache.expire/2","type":"callback"},{"doc":"Flushes the cache and returns the number of evicted keys.Examplesiex&gt; :ok = Enum.each(1..5, &amp;MyCache.put(&amp;1, &amp;1)) iex&gt; MyCache.flush() 5 iex&gt; MyCache.size() 0","ref":"Nebulex.Cache.html#c:flush/0","title":"Nebulex.Cache.flush/0","type":"callback"},{"doc":"Gets a value from Cache where the key matches the given key.Returns nil if no result was found.OptionsSee the &quot;Shared options&quot; section at the module documentation for more options.Exampleiex&gt; MyCache.put(&quot;foo&quot;, &quot;bar&quot;) :ok iex&gt; MyCache.get(&quot;foo&quot;) &quot;bar&quot; iex&gt; MyCache.get(:non_existent_key) nil","ref":"Nebulex.Cache.html#c:get/2","title":"Nebulex.Cache.get/2","type":"callback"},{"doc":"Similar to get/2 but raises KeyError if key is not found.OptionsSee the &quot;Shared options&quot; section at the module documentation for more options.ExampleMyCache.get!(:a)","ref":"Nebulex.Cache.html#c:get!/2","title":"Nebulex.Cache.get!/2","type":"callback"},{"doc":"Returns a map with all the key-value pairs in the Cache where the key is in keys.If keys contains keys that are not in the Cache, they're simply ignored.OptionsSee the &quot;Shared options&quot; section at the module documentation for more options.Exampleiex&gt; MyCache.put_all([a: 1, c: 3]) :ok iex&gt; MyCache.get_all([:a, :b, :c]) %{a: 1, c: 3}","ref":"Nebulex.Cache.html#c:get_all/2","title":"Nebulex.Cache.get_all/2","type":"callback"},{"doc":"Gets the value from key and updates it, all in one pass.fun is called with the current cached value under key (or nil if key hasn't been cached) and must return a two-element tuple: the &quot;get&quot; value (the retrieved value, which can be operated on before being returned) and the new value to be stored under key. fun may also return :pop, which means the current value shall be removed from Cache and returned.The returned value is a tuple with the &quot;get&quot; value returned by fun and the new updated value under key.Options:ttl - Time To Live (TTL) or expiration time in milliseconds for a key (default: :infinity).See the &quot;Shared options&quot; section at the module documentation for more options.Examples# update nonexistent key iex&gt; MyCache.get_and_update(:a, fn current_value -&gt; ...&gt; {current_value, &quot;value!&quot;} ...&gt; end) {nil, &quot;value!&quot;} # update existing key iex&gt; MyCache.get_and_update(:a, fn current_value -&gt; ...&gt; {current_value, &quot;new value!&quot;} ...&gt; end) {&quot;value!&quot;, &quot;new value!&quot;} # pop/remove value if exist iex&gt; MyCache.get_and_update(:a, fn _ -&gt; :pop end) {&quot;new value!&quot;, nil} # pop/remove nonexistent key iex&gt; MyCache.get_and_update(:b, fn _ -&gt; :pop end) {nil, nil}","ref":"Nebulex.Cache.html#c:get_and_update/3","title":"Nebulex.Cache.get_and_update/3","type":"callback"},{"doc":"Returns the atom name or pid of the current cache (based on Ecto dynamic repo).See also put_dynamic_cache/1.","ref":"Nebulex.Cache.html#c:get_dynamic_cache/0","title":"Nebulex.Cache.get_dynamic_cache/0","type":"callback"},{"doc":"Returns whether the given key exists in the Cache.Examplesiex&gt; MyCache.put(:a, 1) :ok iex&gt; MyCache.has_key?(:a) true iex&gt; MyCache.has_key?(:b) false","ref":"Nebulex.Cache.html#c:has_key?/1","title":"Nebulex.Cache.has_key?/1","type":"callback"},{"doc":"Returns true if the current process is inside a transaction.ExamplesMyCache.in_transaction? #=&gt; false MyCache.transaction(fn -&gt; MyCache.in_transaction? #=&gt; true end)","ref":"Nebulex.Cache.html#c:in_transaction?/0","title":"Nebulex.Cache.in_transaction?/0","type":"callback"},{"doc":"Increments or decrements the counter mapped to the given key.If incr &gt;= 0 (positive value) then the current value is incremented by that amount, otherwise, it means the X is a negative value so the current value is decremented by the same amount.Options:ttl - Time To Live (TTL) or expiration time in milliseconds for a key (default: :infinity).See the &quot;Shared options&quot; section at the module documentation for more options.Examplesiex&gt; MyCache.incr(:a) 1 iex&gt; MyCache.incr(:a, 2) 3 iex&gt; MyCache.incr(:a, -1) 2","ref":"Nebulex.Cache.html#c:incr/3","title":"Nebulex.Cache.incr/3","type":"callback"},{"doc":"A callback executed when the cache starts or when configuration is read.","ref":"Nebulex.Cache.html#c:init/1","title":"Nebulex.Cache.init/1","type":"callback"},{"doc":"Loads a dumped cache from the given path.Returns :ok if successful, or {:error, reason} if an error occurs.OptionsSimilar to dump/2, this operation relies entirely on the adapter implementation, therefore, it is recommended to review the documentation of the adapter to be used. Similarly, the built-in adapters inherit the default implementation from Nebulex.Adapter.Persistence, hence, review the available options there.Examples# set some entries iex&gt; entries = for x &lt;- 1..10, into: %{}, do: {x, x} iex&gt; MyCache.set_many(entries) :ok # dump cache to a file iex&gt; MyCache.dump(&quot;my_cache&quot;) :ok # load the cache from a file iex&gt; MyCache.load(&quot;my_cache&quot;) :ok","ref":"Nebulex.Cache.html#c:load/2","title":"Nebulex.Cache.load/2","type":"callback"},{"doc":"Puts the given value under key into the Cache.If key already holds an entry, it is overwritten. Any previous time to live associated with the key is discarded on successful put operation.Options:ttl - Time To Live (TTL) or expiration time in milliseconds for a key (default: :infinity).See the &quot;Shared options&quot; section at the module documentation for more options.Exampleiex&gt; import Nebulex.Time iex&gt; MyCache.put(&quot;foo&quot;, &quot;bar&quot;) :ok # if the value is nil, then it is not stored (operation is skipped) iex&gt; MyCache.put(&quot;foo&quot;, nil) :ok # put key with time-to-live iex&gt; MyCache.put(&quot;foo&quot;, &quot;bar&quot;, ttl: 10_000) :ok # using Nebulex.Time iex&gt; MyCache.put(&quot;foo&quot;, &quot;bar&quot;, ttl: expiry_time(10)) iex&gt; MyCache.put(&quot;foo&quot;, &quot;bar&quot;, ttl: expiry_time(10, :minute)) iex&gt; MyCache.put(&quot;foo&quot;, &quot;bar&quot;, ttl: expiry_time(1, :hour))","ref":"Nebulex.Cache.html#c:put/3","title":"Nebulex.Cache.put/3","type":"callback"},{"doc":"Puts the given entries (key/value pairs) into the Cache. It replaces existing values with new values (just as regular put).Options:ttl - Time To Live (TTL) or expiration time in milliseconds for a key (default: :infinity). It applies to all given entries.See the &quot;Shared options&quot; section at the module documentation for more options.Exampleiex&gt; MyCache.put_all(apples: 3, bananas: 1) :ok iex&gt; MyCache.put_all(%{apples: 2, oranges: 1}, ttl: 10_000) :okIdeally, this operation should be atomic, so all given keys are put at once. But it depends purely on the adapter's implementation and the backend used internally by the adapter. Hence, it is recommended to review the adapter's documentation.","ref":"Nebulex.Cache.html#c:put_all/2","title":"Nebulex.Cache.put_all/2","type":"callback"},{"doc":"Sets the dynamic cache to be used in further commands (based on Ecto dynamic repo).There might be cases where we want to have different cache instances but accessing them through the same cache module. By default, when you call MyApp.Cache.start_link/1, it will start a cache with the name MyApp.Cache. But it is also possible to start multiple caches by using a different name for each of them:MyApp.Cache.start_link(name: :cache1) MyApp.Cache.start_link(name: :cache2, backend: :shards)However, once the cache is started, it is not possible to interact directly with it, since all operations through MyApp.Cache are sent by default to the cache named MyApp.Cache. But you can change the default cache at compile-time:use Nebulex.Cache, default_dynamic_cache: :cache_nameOr anytime at runtime by calling put_dynamic_cache/1:MyApp.Cache.put_dynamic_cache(:another_cache_name)From this moment on, all future commands performed by the current process will run on :another_cache_name.","ref":"Nebulex.Cache.html#c:put_dynamic_cache/1","title":"Nebulex.Cache.put_dynamic_cache/1","type":"callback"},{"doc":"Puts the given value under key into the cache, only if it does not already exist.Returns true if a value was set, otherwise, false is returned.Options:ttl - Time To Live (TTL) or expiration time in milliseconds for a key (default: :infinity).See the &quot;Shared options&quot; section at the module documentation for more options.Exampleiex&gt; MyCache.put_new(&quot;foo&quot;, &quot;bar&quot;) true iex&gt; MyCache.put_new(&quot;foo&quot;, &quot;bar&quot;) false # if the value is nil, it is not stored (operation is skipped) iex&gt; MyCache.put_new(&quot;other&quot;, nil) true","ref":"Nebulex.Cache.html#c:put_new/3","title":"Nebulex.Cache.put_new/3","type":"callback"},{"doc":"Similar to put_new/3 but raises Nebulex.KeyAlreadyExistsError if the key already exists.Options:ttl - Time To Live (TTL) or expiration time in milliseconds for a key (default: :infinity).See the &quot;Shared options&quot; section at the module documentation for more options.ExampleMyCache.put_new!(&quot;foo&quot;, &quot;bar&quot;)","ref":"Nebulex.Cache.html#c:put_new!/3","title":"Nebulex.Cache.put_new!/3","type":"callback"},{"doc":"Puts the given entries (key/value pairs) into the cache. It will not perform any operation at all even if just a single key already exists.Returns true if all entries were successfully set. It returns false if no key was set (at least one key already existed).Options:ttl - Time To Live (TTL) or expiration time in milliseconds for a key (default: :infinity). It applies to all given entries.See the &quot;Shared options&quot; section at the module documentation for more options.Exampleiex&gt; MyCache.put_new_all(apples: 3, bananas: 1) true iex&gt; MyCache.put_new_all(%{apples: 3, oranges: 1}, ttl: 10_000) falseIdeally, this operation should be atomic, so all given keys are put at once. But it depends purely on the adapter's implementation and the backend used internally by the adapter. Hence, it is recommended to review the adapter's documentation.","ref":"Nebulex.Cache.html#c:put_new_all/2","title":"Nebulex.Cache.put_new_all/2","type":"callback"},{"doc":"Alters the entry stored under key, but only if the entry already exists into the Cache.Returns true if a value was set, otherwise, false is returned.Options:ttl - Time To Live (TTL) or expiration time in milliseconds for a key (default: :infinity).See the &quot;Shared options&quot; section at the module documentation for more options.Exampleiex&gt; MyCache.replace(&quot;foo&quot;, &quot;bar&quot;) false iex&gt; MyCache.put_new(&quot;foo&quot;, &quot;bar&quot;) true iex&gt; MyCache.replace(&quot;foo&quot;, &quot;bar2&quot;) true # update current value and TTL iex&gt; MyCache.replace(&quot;foo&quot;, &quot;bar3&quot;, ttl: 10_000) true","ref":"Nebulex.Cache.html#c:replace/3","title":"Nebulex.Cache.replace/3","type":"callback"},{"doc":"Similar to replace/3 but raises KeyError if key is not found.Options:ttl - Time To Live (TTL) or expiration time in milliseconds for a key (default: :infinity).See the &quot;Shared options&quot; section at the module documentation for more options.ExampleMyCache.replace!(&quot;foo&quot;, &quot;bar&quot;)","ref":"Nebulex.Cache.html#c:replace!/3","title":"Nebulex.Cache.replace!/3","type":"callback"},{"doc":"Returns the total number of cached entries.Examplesiex&gt; :ok = Enum.each(1..10, &amp;MyCache.put(&amp;1, &amp;1)) iex&gt; MyCache.size() 10 iex&gt; :ok = Enum.each(1..5, &amp;MyCache.delete(&amp;1)) iex&gt; MyCache.size() 5","ref":"Nebulex.Cache.html#c:size/0","title":"Nebulex.Cache.size/0","type":"callback"},{"doc":"Starts a supervision and return {:ok, pid} or just :ok if nothing needs to be done.Returns {:error, {:already_started, pid}} if the cache is already started or {:error, term} in case anything else goes wrong.OptionsSee the configuration in the moduledoc for options shared between adapters, for adapter-specific configuration see the adapter's documentation.","ref":"Nebulex.Cache.html#c:start_link/1","title":"Nebulex.Cache.start_link/1","type":"callback"},{"doc":"Shuts down the cache.","ref":"Nebulex.Cache.html#c:stop/1","title":"Nebulex.Cache.stop/1","type":"callback"},{"doc":"Similar to all/2 but returns a lazy enumerable that emits all entries from the cache matching the given query.May raise Nebulex.QueryError if query validation fails.Options:return - Tells the query what to return from the matched entries. The possible values are: :key, :value, and :entry ({key, value} pairs). Defaults to :key. This option is supported by the build-in adapters, but it is recommended to check the adapter's documentation to confirm its compatibility with this option.:page_size - Positive integer (&gt;= 1) that defines the page size for the stream (defaults to 10).See the &quot;Shared options&quot; section at the module documentation for more options.Examples# set some entries iex&gt; :ok = Enum.each(1..5, &amp;MyCache.put(&amp;1, &amp;1 * 2)) # stream all (with default params) iex&gt; MyCache.stream() |&gt; Enum.to_list() [1, 2, 3, 4, 5] # stream all entries and return values iex&gt; MyCache.stream(nil, return: :value, page_size: 3) |&gt; Enum.to_list() [2, 4, 6, 8, 10] # additional built-in queries for Nebulex.Adapters.Local adapter unexpired_stream = MyCache.stream(:unexpired) expired_stream = MyCache.stream(:expired) # if we are using Nebulex.Adapters.Local adapter, the stored entry tuple # {:entry, key, value, version, expire_at}, then the match spec could be # something like: iex&gt; spec = [{{:entry, :&quot;$1&quot;, :&quot;$2&quot;, :_, :_}, [{:&gt;, :&quot;$2&quot;, 5}], [{{:&quot;$1&quot;, :&quot;$2&quot;}}]}] iex&gt; MyCache.stream(spec, page_size: 3) |&gt; Enum.to_list() [{3, 6}, {4, 8}, {5, 10}] # the same previous query but using Ex2ms iex&gt; import Ex2ms Ex2ms iex&gt; spec = ...&gt; fun do ...&gt; {_, key, value, _, _} when value &gt; 5 -&gt; {key, value} ...&gt; end iex&gt; spec |&gt; MyCache.stream(page_size: 3) |&gt; Enum.to_list() [{3, 6}, {4, 8}, {5, 10}]","ref":"Nebulex.Cache.html#c:stream/2","title":"Nebulex.Cache.stream/2","type":"callback"},{"doc":"Returns and removes the value associated with key in the Cache. If the key does not exist, then nil is returned.OptionsSee the &quot;Shared options&quot; section at the module documentation for more options.Examplesiex&gt; MyCache.put(:a, 1) :ok iex&gt; MyCache.take(:a) 1 iex&gt; MyCache.take(:a) nil","ref":"Nebulex.Cache.html#c:take/2","title":"Nebulex.Cache.take/2","type":"callback"},{"doc":"Similar to take/2 but raises KeyError if key is not found.OptionsSee the &quot;Shared options&quot; section at the module documentation for more options.ExampleMyCache.take!(:a)","ref":"Nebulex.Cache.html#c:take!/2","title":"Nebulex.Cache.take!/2","type":"callback"},{"doc":"Returns true if the given key exists and the last access time was successfully updated, otherwise, false is returned.Examplesiex&gt; MyCache.put(:a, 1) :ok iex&gt; MyCache.touch(:a) true iex&gt; MyCache.ttl(:b) false","ref":"Nebulex.Cache.html#c:touch/1","title":"Nebulex.Cache.touch/1","type":"callback"},{"doc":"Runs the given function inside a transaction.A successful transaction returns the value returned by the function.OptionsSee the &quot;Shared options&quot; section at the module documentation for more options.ExamplesMyCache.transaction fn -&gt; alice = MyCache.get(:alice) bob = MyCache.get(:bob) MyCache.put(:alice, %{alice | balance: alice.balance + 100}) MyCache.put(:bob, %{bob | balance: bob.balance + 100}) end # locking only the involved key (recommended): MyCache.transaction [keys: [:alice, :bob]], fn -&gt; alice = MyCache.get(:alice) bob = MyCache.get(:bob) MyCache.put(:alice, %{alice | balance: alice.balance + 100}) MyCache.put(:bob, %{bob | balance: bob.balance + 100}) end","ref":"Nebulex.Cache.html#c:transaction/2","title":"Nebulex.Cache.transaction/2","type":"callback"},{"doc":"Returns the remaining time-to-live for the given key. If the key does not exist, then nil is returned.Examplesiex&gt; MyCache.put(:a, 1, ttl: 5000) :ok iex&gt; MyCache.put(:b, 2) :ok iex&gt; MyCache.ttl(:a) _remaining_ttl iex&gt; MyCache.ttl(:b) :infinity iex&gt; MyCache.ttl(:c) nil","ref":"Nebulex.Cache.html#c:ttl/1","title":"Nebulex.Cache.ttl/1","type":"callback"},{"doc":"Updates the cached key with the given function.If key is present in Cache with value value, fun is invoked with argument value and its result is used as the new value of key.If key is not present in Cache, initial is inserted as the value of key. The initial value will not be passed through the update function.Options:ttl - Time To Live (TTL) or expiration time in milliseconds for a key (default: :infinity).See the &quot;Shared options&quot; section at the module documentation for more options.Examplesiex&gt; MyCache.update(:a, 1, &amp;(&amp;1 * 2)) 1 iex&gt; MyCache.update(:a, 1, &amp;(&amp;1 * 2)) 2","ref":"Nebulex.Cache.html#c:update/4","title":"Nebulex.Cache.update/4","type":"callback"},{"doc":"Cache entries","ref":"Nebulex.Cache.html#t:entries/0","title":"Nebulex.Cache.entries/0","type":"type"},{"doc":"Cache entry key","ref":"Nebulex.Cache.html#t:key/0","title":"Nebulex.Cache.key/0","type":"type"},{"doc":"Cache action options","ref":"Nebulex.Cache.html#t:opts/0","title":"Nebulex.Cache.opts/0","type":"type"},{"doc":"","ref":"Nebulex.Cache.html#t:t/0","title":"Nebulex.Cache.t/0","type":"type"},{"doc":"Cache entry value","ref":"Nebulex.Cache.html#t:value/0","title":"Nebulex.Cache.value/0","type":"type"},{"doc":"This module defines the supported built-in stats.By default, each adapter is responsible for providing stats support. However, Nebulex suggests supporting the built-in stats described in this module, which are also supported by the built-in adapters.UsageFirst of all, we define a cache:defmodule MyApp.Cache do use Nebulex.Cache, otp_app: :nebulex, adapter: Nebulex.Adapters.Local endThen we configure it enabling the stats, like so:config :my_app, MyApp.Cache, stats: true, gc_interval: 86_400_000, #=&gt; 1 day max_size: 200_000, gc_cleanup_min_timeout: 10_000, gc_cleanup_max_timeout: 900_000Remember to add the cache on your application's supervision tree.Since we are using a built-in adapter and the stats have been enabled (stats: true), the stat counters will be automatically fed by the adapter.You can ask for the current stats values at any time by calling:Nebulex.Cache.Stats.info(MyApp.Cache)Using stats helpersYou can inject the stats helpers in the cache like this:defmodule MyApp.Cache do use Nebulex.Cache, otp_app: :nebulex, adapter: Nebulex.Adapters.Local # Use stats helpers use Nebulex.Cache.Stats endRetrieving stats infoMyApp.Cache.stats_info()By calling this injected helper, the function Nebulex.Cache.Stats.info/1 is called under-the-hood, but the cache name is resolved automatically.Dispatching telemetry eventsMyApp.Cache.dispatch_stats() MyApp.Cache.dispatch_stats(event_prefix: [:my_cache, :stats])By calling this injected helper, the function Nebulex.Cache.Stats.dispatch/2 is called under-the-hood, but the cache name is resolved automatically.Telemetry eventsIntegrating telemetry is very easy since with the helper function MyApp.Cache.dispatch_stats/1 (via the __using__ macro) described previously you can emit telemetry events with the current stats at any time. What we need to resolve is, how to make it in such a way that every X period of time the stats are emitted automatically.To do so, we can use :telemetry_poller and define a custom measurement::telemetry_poller.start_link( measurements: [ {MyApp.Cache, :dispatch_stats, []}, ], # configure sampling period - default is :timer.seconds(5) period: :timer.seconds(10), name: :my_cache_stats_poller )Or you can also start the :telemetry_poller process along with your application supervision tree, like so:def start(_type, _args) do my_cache_stats_poller_opts = [ measurements: [ {MyApp.Cache, :dispatch_stats, []}, ], period: :timer.seconds(10), name: :my_cache_stats_poller ] children = [ {MyApp.Cache, []}, {:telemetry_poller, my_cache_stats_poller_opts} ] opts = [strategy: :one_for_one, name: MyApp.Supervisor] Supervisor.start_link(children, opts) endSee Nebulex Telemetry Guide.","ref":"Nebulex.Cache.Stats.html","title":"Nebulex.Cache.Stats","type":"module"},{"doc":"Emits a telemetry event when called with the current stats count.The :measurements map will include the current count for each stat::hits - Current hits count.:misses - Current misses count.:writes - Current writes count.:evictions - Current evictions count.:expirations - Current expirations count.The telemetry :metadata map will include the following fields::cache - The cache module, or the name (if an explicit name has been given to the cache).Additionally, you can add your own metadata fields by given the option :metadata.Options:event_prefix – The prefix of the telemetry event. Defaults to [:nebulex, :cache].:metadata – A map with additional metadata fields. Defaults to %{}.Examplesiex&gt; Nebulex.Cache.Stats.dispatch(MyCache) :ok iex&gt; Nebulex.Cache.Stats.dispatch( ...&gt; MyCache, ...&gt; event_prefix: [:my_cache], ...&gt; metadata: %{tag: &quot;tag1&quot;} ...&gt; ) :ok","ref":"Nebulex.Cache.Stats.html#dispatch/2","title":"Nebulex.Cache.Stats.dispatch/2","type":"function"},{"doc":"Increments the counter's stat stat by the given incr value.ExamplesNebulex.Cache.Stats.incr(stat_counter, :hits) Nebulex.Cache.Stats.incr(stat_counter, :writes, 10)NOTE: This function is normally called by the adapter in case it supports the Nebulex suggested stats; the adapter should feed Nebulex.Cache.Stats.t() counters.See built-in adapters for more information about the usage.","ref":"Nebulex.Cache.Stats.html#incr/3","title":"Nebulex.Cache.Stats.incr/3","type":"function"},{"doc":"Returns the struct Nebulex.Cache.Stats with the current stats values for the given cache name or counter reference. Normally, the cache name is passed so that the counter reference is retrieved and handled internally.Returns nil if the stats are disabled or if the adapter doesn't support this feature.Exampleiex&gt; Nebulex.Cache.Stats.info(MyCache) %Nebulex.Cache.Stats{ evictions: 0, expirations: 0, hits: 0, misses: 0, writes: 0 }","ref":"Nebulex.Cache.Stats.html#info/1","title":"Nebulex.Cache.Stats.info/1","type":"function"},{"doc":"Initializes the Erlang's counter to be used for the calling Cache and feed the stat values; see the module documentation for more information about the supported stats.Returns nil is the option :stats is set to false or it is not set at all; the stats will be skipped.ExampleNebulex.Cache.Stats.init(opts)NOTE: This function is normally called by the adapter in case it supports the Nebulex suggested stats; the adapter should feed Nebulex.Cache.Stats.t() counters.See built-in adapters for more information about the usage.","ref":"Nebulex.Cache.Stats.html#init/1","title":"Nebulex.Cache.Stats.init/1","type":"function"},{"doc":"","ref":"Nebulex.Cache.Stats.html#t:stat/0","title":"Nebulex.Cache.Stats.stat/0","type":"type"},{"doc":"","ref":"Nebulex.Cache.Stats.html#t:t/0","title":"Nebulex.Cache.Stats.t/0","type":"type"},{"doc":"Declarative annotation-based caching via function decorators.For caching declaration, the abstraction provides three Elixir function decorators: cacheable, cache_evict, and cache_put, which allow functions to trigger cache population or cache eviction. Let us take a closer look at each annotation.Inspired by Spring Cache Abstraction.cacheable decoratorAs the name implies, cacheable is used to demarcate functions that are cacheable - that is, functions for whom the result is stored into the cache so, on subsequent invocations (with the same arguments), the value in the cache is returned without having to actually execute the function. In its simplest form, the decorator/annotation declaration requires the name of the cache associated with the annotated function:@decorate cacheable(cache: Cache) def get_account(id) do # the logic for retrieving the account ... endIn the snippet above, the function get_account/1 is associated with the cache named Cache. Each time the function is called, the cache is checked to see whether the invocation has been already executed and does not have to be repeated.Default Key GenerationSince caches are essentially key-value stores, each invocation of a cached function needs to be translated into a suitable key for cache access. Out of the box, the caching abstraction uses a simple key-generator based on the following algorithm: :erlang.phash2({module, func_name}).Custom Key Generation DeclarationSince caching is generic, it is quite likely the target functions have various signatures that cannot be simply mapped on top of the cache structure. This tends to become obvious when the target function has multiple arguments out of which only some are suitable for caching (while the rest are used only by the function logic). For example:@decorate cacheable(cache: Cache) def get_account(email, include_users?) do # the logic for retrieving the account ... endAt first glance, while the boolean argument influences the way the account is found, it is no use for the cache.For such cases, the cacheable decorator allows the user to specify the key explicitly based on the function attributes.@decorate cacheable(cache: Cache, key: {Account, email}) def get_account(email, include_users?) do # the logic for retrieving the account ... end @decorate cacheable(cache: Cache, key: {Account, user.account_id}) def get_user_account(%User{} = user) do # the logic for retrieving the account ... endIt is also possible passing options to the cache, like so:@decorate cacheable(cache: Cache, key: {Account, email}, opts: [ttl: 300_000]) def get_account(email, include_users?) do # the logic for retrieving the account ... endSee the &quot;Shared Options&quot; section below.Functions with multiple clausesSince decorator lib is used, it is important to be aware of the recommendations, warns, limitations, and so on. In this case, for functions with multiple clauses the general advice is to create an empty function head, and call the decorator on that head, like so:@decorate cacheable(cache: Cache, key: email) def get_account(email \\\\ nil) def get_account(nil), do: nil def get_account(email) do # the logic for retrieving the account ... endcache_put decoratorFor cases where the cache needs to be updated without interfering with the function execution, one can use the cache_put decorator. That is, the method will always be executed and its result placed into the cache (according to the cache_put options). It supports the same options as cacheable.@decorate cache_put(cache: Cache, key: {Account, acct.email}) def update_account(%Account{} = acct, attrs) do # the logic for updating the account ... endNote that using cache_put and cacheable annotations on the same function is generally discouraged because they have different behaviors. While the latter causes the method execution to be skipped by using the cache, the former forces the execution in order to execute a cache update. This leads to unexpected behavior and with the exception of specific corner-cases (such as decorators having conditions that exclude them from each other), such declarations should be avoided.cache_evict decoratorThe cache abstraction allows not just the population of a cache store but also eviction. This process is useful for removing stale or unused data from the cache. Opposed to cacheable, the decorator cache_evict demarcates functions that perform cache eviction, which are functions that act as triggers for removing data from the cache. The cache_evict decorator not only allows a key to be specified, but also a set of keys. Besides, extra options likeall_entries which indicates whether a cache-wide eviction needs to be performed rather than just an entry one (based on the key or keys):@decorate cache_evict(cache: Cache, key: {Account, email}) def delete_account_by email(email) do # the logic for deleting the account ... end @decorate cacheable(cache: Cache, keys: [{Account, acct.id}, {Account, acct.email}]) def delete_account(%Account{} = acct) do # the logic for deleting the account ... end @decorate cacheable(cache: Cache, all_entries: true) def delete_all_accounts do # the logic for deleting all the accounts ... endThe option all_entries: comes in handy when an entire cache region needs to be cleared out - rather than evicting each entry (which would take a long time since it is inefficient), all the entries are removed in one operation as shown above.Shared OptionsAll three cache annotations explained previously accept the following options::cache - Defines what cache to use (required). Raises ArgumentError if the option is not present.:key - Defines the cache access key (optional). If this option is not present, a default key is generated by hashing the tuple {module, fun_name}; the first element is the caller module and the second one the function name (:erlang.phash2({module, fun})).:opts - Defines the cache options that will be passed as argument to the invoked cache function (optional).:match - Match function (term -&gt; boolean | {true, term}) (optional). This function is for matching and decide whether or not the code-block evaluation result is cached or not. If true the code-block evaluation result is cached as it is (the default). If {true, value} is returned, then the value is what is cached (useful to control what to cache). Otherwise, none result is stored in the cache.Putting all togetherSupposing we are using Ecto and we want to define some cacheable functions within the context MyApp.Accounts:# The config config :my_app, MyApp.Cache, gc_interval: 86_400_000, #=&gt; 1 day backend: :shards # The Cache defmodule MyApp.Cache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local end # Some Ecto schema defmodule MyApp.Accounts.User do use Ecto.Schema schema &quot;users&quot; do field(:username, :string) field(:password, :string) field(:role, :string) end def changeset(user, attrs) do user |&gt; cast(attrs, [:username, :password, :role]) |&gt; validate_required([:username, :password, :role]) end end # Accounts context defmodule MyApp.Accounts do use Nebulex.Caching alias MyApp.Accounts.User alias MyApp.{Cache, Repo} @ttl Nebulex.Time.expiry_time(1, :hour) @decorate cacheable(cache: Cache, key: {User, id}, opts: [ttl: @ttl]) def get_user!(id) do Repo.get!(User, id) end @decorate cacheable(cache: Cache, key: {User, username}, opts: [ttl: @ttl]) def get_user_by_username(username) do Repo.get_by(User, [username: username]) end @decorate cache_put( cache: Cache, keys: [{User, usr.id}, {User, usr.username}], match: &amp;match_update/1 ) def update_user(%User{} = usr, attrs) do usr |&gt; User.changeset(attrs) |&gt; Repo.update() end defp match_update({:ok, usr}), do: {true, usr} defp match_update({:error, _}), do: false @decorate cache_evict(cache: Cache, keys: [{User, usr.id}, {User, usr.username}]) def delete_user(%User{} = usr) do Repo.delete(usr) end def create_user(attrs \\\\ %{}) do %User{} |&gt; User.changeset(attrs) |&gt; Repo.insert() end endSee Cache Usage Patters Guide.","ref":"Nebulex.Caching.html","title":"Nebulex.Caching","type":"module"},{"doc":"","ref":"Nebulex.Caching.html#cache_evict/1","title":"Nebulex.Caching.cache_evict/1","type":"macro"},{"doc":"Provides a way of annotating functions to be evicted (eviction aspect).On function's completion, the given key or keys (depends on the :key and :keys options) are deleted from the cache.Options:keys - Defines the set of keys to be evicted from cache on function completion.:all_entries - Defines if all entries must be removed on function completion. Defaults to false.See the &quot;Shared options&quot; section at the module documentation.Examplesdefmodule MyApp.Example do use Nebulex.Caching alias MyApp.Cache @decorate cache_evict(cache: Cache, key: id) def delete(id) do # your logic (maybe write/delete data to the SoR) end @decorate cache_evict(cache: Cache, keys: [object.name, object.id]) def delete_object(object) do # your logic (maybe write/delete data to the SoR) end @decorate cache_evict(cache: Cache, all_entries: true) def delete_all do # your logic (maybe write/delete data to the SoR) end endThe Write-through pattern is supported by this decorator. Your function provides the logic to write data to the system-of-record (SoR) and the rest is provided by the decorator under-the-hood. But in contrast with update decorator, when the data is written to the SoR, the key for that value is deleted from cache instead of updated.","ref":"Nebulex.Caching.html#cache_evict/3","title":"Nebulex.Caching.cache_evict/3","type":"function"},{"doc":"","ref":"Nebulex.Caching.html#cache_put/1","title":"Nebulex.Caching.cache_put/1","type":"macro"},{"doc":"Provides a way of annotating functions to be evicted; but updating the cached key instead of deleting it.The content of the cache is updated without interfering with the function execution. That is, the method would always be executed and the result cached.The difference between cacheable/3 and cache_put/3 is that cacheable/3 will skip running the function if the key exists in the cache, whereas cache_put/3 will actually run the function and then put the result in the cache.OptionsSee the &quot;Shared options&quot; section at the module documentation.Examplesdefmodule MyApp.Example do use Nebulex.Caching alias MyApp.Cache @ttl Nebulex.Time.expiry_time(1, :hour) @decorate cache_put(cache: Cache, key: id, opts: [ttl: @ttl]) def update!(id, attrs \\ %{}) do # your logic (maybe write data to the SoR) end @decorate cache_put(cache: Cache, key: id, match: &amp;match_fun/1, opts: [ttl: @ttl]) def update(id, attrs \\ %{}) do # your logic (maybe write data to the SoR) end defp match_fun({:ok, updated}), do: {true, updated} defp match_fun({:error, _}), do: false endThe Write-through pattern is supported by this decorator. Your function provides the logic to write data to the system-of-record (SoR) and the rest is provided by the decorator under-the-hood.","ref":"Nebulex.Caching.html#cache_put/3","title":"Nebulex.Caching.cache_put/3","type":"function"},{"doc":"","ref":"Nebulex.Caching.html#cacheable/1","title":"Nebulex.Caching.cacheable/1","type":"macro"},{"doc":"Provides a way of annotating functions to be cached (cacheable aspect).The returned value by the code block is cached if it doesn't exist already in cache, otherwise, it is returned directly from cache and the code block is not executed.OptionsSee the &quot;Shared options&quot; section at the module documentation.Examplesdefmodule MyApp.Example do use Nebulex.Caching alias MyApp.Cache @ttl Nebulex.Time.expiry_time(1, :hour) @decorate cacheable(cache: Cache, key: name) def get_by_name(name, age) do # your logic (maybe the loader to retrieve the value from the SoR) end @decorate cacheable(cache: Cache, key: age, opts: [ttl: @ttl]) def get_by_age(age) do # your logic (maybe the loader to retrieve the value from the SoR) end @decorate cacheable(cache: Cache, key: clauses, match: &amp;match_fun/1) def all(clauses) do # your logic (maybe the loader to retrieve the value from the SoR) end defp match_fun([]), do: false defp match_fun(_), do: true endThe Read-through pattern is supported by this decorator. The loader to retrieve the value from the system-of-record (SoR) is your function's logic and the rest is provided by the macro under-the-hood.","ref":"Nebulex.Caching.html#cacheable/3","title":"Nebulex.Caching.cacheable/3","type":"function"},{"doc":"This function is for internal purposes.NOTE: Workaround to avoid dialyzer warnings when using declarative annotation-based caching via decorators.","ref":"Nebulex.Caching.html#eval_match/5","title":"Nebulex.Caching.eval_match/5","type":"function"},{"doc":"Defines a Cache Entry.This is the structure used by the caches for representing cache entries.","ref":"Nebulex.Entry.html","title":"Nebulex.Entry","type":"module"},{"doc":"Decodes a previously encoded entry.Exampleiex&gt; %Nebulex.Entry{} ...&gt; |&gt; Nebulex.Entry.encode() ...&gt; |&gt; Nebulex.Entry.decode() _decoded_entry","ref":"Nebulex.Entry.html#decode/2","title":"Nebulex.Entry.decode/2","type":"function"},{"doc":"Encodes a cache entry.Exampleiex&gt; Nebulex.Entry.encode(%Nebulex.Entry{}) _encoded_entry","ref":"Nebulex.Entry.html#encode/2","title":"Nebulex.Entry.encode/2","type":"function"},{"doc":"Returns whether the given entry has expired or not.Exampleiex&gt; Nebulex.Entry.expired?(%Nebulex.Entry{}) false","ref":"Nebulex.Entry.html#expired?/1","title":"Nebulex.Entry.expired?/1","type":"function"},{"doc":"Returns the remaining time-to-live.Exampleiex&gt; Nebulex.Entry.ttl(%Nebulex.Entry{}) :infinity","ref":"Nebulex.Entry.html#ttl/1","title":"Nebulex.Entry.ttl/1","type":"function"},{"doc":"","ref":"Nebulex.Entry.html#t:t/0","title":"Nebulex.Entry.t/0","type":"type"},{"doc":"Pre/Post HooksSince v2.0.0, pre/post hooks are not supported and/or handled by Nebulex itself. Hooks feature is not a common use-case and also it is something that can be be easily implemented on top of the Cache at the application level.Nevertheless, to keep backward compatibility somehow, Nebulex provides the next decorators for implementing pre/post hooks very easily.before decoratorThe before decorator is declared for performing a hook action or callback before the annotated function is executed.@decorate before(fn %Nebulex.Hook{} = hook -&gt; inspect(hook) end) def some_fun(var) do # logic ... endafter_return decoratorThe after_return decorator is declared for performing a hook action or callback after the annotated function is executed and its return is passed through the return: attribute.@decorate after_return(&amp;inspect(&amp;1.return)) def some_fun(var) do # logic ... endaround decoratorThe final kind of hook is around decorator. The around decorator runs &quot;around&quot; the annotated function execution. It has the opportunity to do work both before and after the function executes. This means the given hook function is invoked twice, before and after the code-block is evaluated.@decorate around(&amp;inspect(&amp;1.step)) def some_fun(var) do # logic ... endPutting all togetherSuppose we want to track all cache calls (before and after they are called) by logging them (including the execution time). In this case, we need to provide a pre/post hook to log these calls.First of all, we have to create a module implementing the hook function:defmodule MyApp.Tracker do use GenServer alias Nebulex.Hook require Logger @actions [:get, :put] ## API def start_link(opts \\\\ []) do GenServer.start_link(__MODULE__, opts, name: __MODULE__) end def track(%Hook{step: :before, name: name}) when name in @actions do System.system_time(:microsecond) end def track(%Hook{step: :after_return, name: name} = event) when name in @actions do GenServer.cast(__MODULE__, {:track, event}) end def track(hook), do: hook ## GenServer Callbacks @impl true def init(_opts) do {:ok, %{}} end @impl true def handle_cast({:track, %Hook{acc: start} = hook}, state) do diff = System.system_time(:microsecond) - start Logger.info(&quot;#=&gt; \#{hook.module}.\#{hook.name}/\#{hook.arity}, Duration: \#{diff}&quot;) {:noreply, state} end endAnd then, in the Cache:defmodule MyApp.Cache do use Nebulex.Hook @decorate_all around(&amp;MyApp.Tracker.track/1) use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local endTry it out:iex&gt; MyApp.Cache.put 1, 1 10:19:47.736 [info] Elixir.MyApp.Cache.put/3, Duration: 27 iex&gt; MyApp.Cache.get 1 10:20:14.941 [info] Elixir.MyApp.Cache.get/2, Duration: 11","ref":"Nebulex.Hook.html","title":"Nebulex.Hook","type":"module"},{"doc":"","ref":"Nebulex.Hook.html#after_return/1","title":"Nebulex.Hook.after_return/1","type":"macro"},{"doc":"After-return decorator.Intercepts any call to the annotated function and calls the given fun after the logic is executed, and the returned result is passed through the return: attribute.Exampledefmodule MyApp.Example do use Nebulex.Hook @decorate after_return(&amp;inspect(&amp;1)) def some_fun(var) do # logic ... end end","ref":"Nebulex.Hook.html#after_return/3","title":"Nebulex.Hook.after_return/3","type":"function"},{"doc":"","ref":"Nebulex.Hook.html#around/1","title":"Nebulex.Hook.around/1","type":"macro"},{"doc":"Around decorator.Intercepts any call to the annotated function and calls the given fun before and after the logic is executed. The result of the first call to the hook function is passed through the acc: attribute, so it can be used in the next call (after return). Finally, as the after_return decorator, the returned code-block evaluation is passed through the return: attribute.Exampledefmodule MyApp.Profiling do alias Nebulex.Hook def prof(%Hook{step: :before}) do System.system_time(:microsecond) end def prof(%Hook{step: :after_return, acc: start} = hook) do :telemetry.execute( [:my_app, :profiling], %{duration: System.system_time(:microsecond) - start}, %{module: hook.module, name: hook.name} ) end end defmodule MyApp.Example do use Nebulex.Hook @decorate around(&amp;MyApp.Profiling.prof/1) def some_fun(var) do # logic ... end end","ref":"Nebulex.Hook.html#around/3","title":"Nebulex.Hook.around/3","type":"function"},{"doc":"","ref":"Nebulex.Hook.html#before/1","title":"Nebulex.Hook.before/1","type":"macro"},{"doc":"Before decorator.Intercepts any call to the annotated function and calls the given fun before the logic is executed.Exampledefmodule MyApp.Example do use Nebulex.Hook @decorate before(&amp;inspect(&amp;1)) def some_fun(var) do # logic ... end end","ref":"Nebulex.Hook.html#before/3","title":"Nebulex.Hook.before/3","type":"function"},{"doc":"This function is for internal purposes.","ref":"Nebulex.Hook.html#eval_hook/3","title":"Nebulex.Hook.eval_hook/3","type":"function"},{"doc":"","ref":"Nebulex.Hook.html#t:hook_fun/0","title":"Nebulex.Hook.hook_fun/0","type":"type"},{"doc":"","ref":"Nebulex.Hook.html#t:t/0","title":"Nebulex.Hook.t/0","type":"type"},{"doc":"RPC utilities for distributed task execution.This module uses supervised tasks underneath Task.Supervisor.","ref":"Nebulex.RPC.html","title":"Nebulex.RPC","type":"module"},{"doc":"Evaluates apply(mod, fun, args) on node node and returns the corresponding evaluation result, or {:badrpc, reason} if the call fails.A timeout, in milliseconds or :infinity, can be given with a default value of 5000. It uses Task.await/2 internally.Exampleiex&gt; Nebulex.RPC.call(:my_task_sup, :node1, Kernel, :to_string, [1]) &quot;1&quot;","ref":"Nebulex.RPC.html#call/6","title":"Nebulex.RPC.call/6","type":"function"},{"doc":"In contrast to a regular single-node RPC, a multicall is an RPC that is sent concurrently from one client to multiple servers. The function evaluates apply(mod, fun, args) on each node_group entry and collects the answers. Then, evaluates the reducer function (set in the opts) on each answer.This function is similar to :rpc.multicall/5.Options:timeout - A timeout, in milliseconds or :infinity, can be given with a default value of 5000. It uses Task.yield_many/2 internally.:reducer - Reducer function to be executed on each collected result. (check out reducer type).Exampleiex&gt; Nebulex.RPC.multi_call( ...&gt; :my_task_sup, ...&gt; %{ ...&gt; node1: {Kernel, :to_string, [1]}, ...&gt; node2: {Kernel, :to_string, [2]} ...&gt; }, ...&gt; timeout: 10_000, ...&gt; reducer: { ...&gt; [], ...&gt; fn ...&gt; {:ok, res}, _node_callback, acc -&gt; ...&gt; [res | acc] ...&gt; ...&gt; {:error, _}, _node_callback, acc -&gt; ...&gt; acc ...&gt; end ...&gt; } ...&gt; ) [&quot;1&quot;, &quot;2&quot;]","ref":"Nebulex.RPC.html#multi_call/3","title":"Nebulex.RPC.multi_call/3","type":"function"},{"doc":"Similar to multi_call/3 but the same node_callback (given by module, fun, args) is executed on all nodes; Internally it creates a node_group with the same node_callback for each node.OptionsSame options as multi_call/3.Exampleiex&gt; Nebulex.RPC.multi_call( ...&gt; :my_task_sup, ...&gt; [:node1, :node2], ...&gt; Kernel, ...&gt; :to_string, ...&gt; [1], ...&gt; timeout: 5000, ...&gt; reducer: { ...&gt; [], ...&gt; fn ...&gt; {:ok, res}, _node_callback, acc -&gt; ...&gt; [res | acc] ...&gt; ...&gt; {:error, _}, _node_callback, acc -&gt; ...&gt; acc ...&gt; end ...&gt; } ...&gt; ) [&quot;1&quot;, &quot;1&quot;]","ref":"Nebulex.RPC.html#multi_call/6","title":"Nebulex.RPC.multi_call/6","type":"function"},{"doc":"Task callback","ref":"Nebulex.RPC.html#t:callback/0","title":"Nebulex.RPC.callback/0","type":"type"},{"doc":"Group entry: node -&gt; callback","ref":"Nebulex.RPC.html#t:node_callback/0","title":"Nebulex.RPC.node_callback/0","type":"type"},{"doc":"Node group","ref":"Nebulex.RPC.html#t:node_group/0","title":"Nebulex.RPC.node_group/0","type":"type"},{"doc":"Reducer spec","ref":"Nebulex.RPC.html#t:reducer/0","title":"Nebulex.RPC.reducer/0","type":"type"},{"doc":"Time utilities.","ref":"Nebulex.Time.html","title":"Nebulex.Time","type":"module"},{"doc":"","ref":"Nebulex.Time.html#do_expiry_time/2","title":"Nebulex.Time.do_expiry_time/2","type":"function"},{"doc":"Returns the equivalent expiration time in milliseconds for the given time.Exampleiex&gt; Nebulex.Time.expiry_time(10, :second) 10000 iex&gt; Nebulex.Time.expiry_time(10, :minute) 600000 iex&gt; Nebulex.Time.expiry_time(1, :hour) 3600000 iex&gt; Nebulex.Time.expiry_time(1, :day) 86400000","ref":"Nebulex.Time.html#expiry_time/2","title":"Nebulex.Time.expiry_time/2","type":"function"},{"doc":"Returns the current system time in the given time unit.The unit is set to :millisecond by default.Examplesiex&gt; Nebulex.Time.now() _milliseconds","ref":"Nebulex.Time.html#now/1","title":"Nebulex.Time.now/1","type":"function"},{"doc":"Returns true if the given timeout is a valid timeout; otherwise, false is returned.Valid timeout: :infinity | non_neg_integer().Examplesiex&gt; Nebulex.Time.timeout?(1) true iex&gt; Nebulex.Time.timeout?(:infinity) true iex&gt; Nebulex.Time.timeout?(-1) false","ref":"Nebulex.Time.html#timeout?/1","title":"Nebulex.Time.timeout?/1","type":"function"},{"doc":"Raised at runtime when a key already exists in cache.","ref":"Nebulex.KeyAlreadyExistsError.html","title":"Nebulex.KeyAlreadyExistsError","type":"exception"},{"doc":"Raised at runtime when the query is invalid.","ref":"Nebulex.QueryError.html","title":"Nebulex.QueryError","type":"exception"},{"doc":"Raised at runtime when a RPC multi_call error occurs.","ref":"Nebulex.RPCMultiCallError.html","title":"Nebulex.RPCMultiCallError","type":"exception"},{"doc":"Getting StartedThis guide is an introduction to Nebulex, a local and distributed caching toolkif for Elixir. Nebulex API is pretty much inspired by Ecto, taking advantage of its simplicity, flexibility and pluggable architecture. In the same way as Ecto, developers can provide their own cache (adapter) implementations.In this guide, we're going to learn some basics about Nebulex, such as insert, retrieve and destroy cache entries (key/value pairs).","ref":"getting-started.html","title":"Getting Started","type":"extras"},{"doc":"Let's start creating a new Elixir application by running this command:mix new blog --supThe --sup option ensures that this application has a supervision tree, which will be needed by Nebulex later on.To add Nebulex to this application, there are a few steps that we need to take.The first step will be adding Nebulex to our mix.exs file, which we'll do by changing the deps definition in that file to this:defp deps do [ {:nebulex, &quot;~&gt; 2.0&quot;}, {:shards, &quot;~&gt; 0.6&quot;}, #=&gt; Since we will use :shards as backend {:decorator, &quot;~&gt; 1.3&quot;} #=&gt; If you want to use Caching Annotations ] endIn order to give more flexibility and loading only needed dependencies, Nebulex makes all its dependencies as optional. For example:For intensive workloads, we may want to use :shards as the backend for the local adapter and having partitioned tables. In such a case, you have to add :shards to the dependency list.For enabling the usage of declarative annotation-based caching via decorators, you have to add :decorator to the dependency list.Also, all the external adapters have to be added as a dependency as well.To install these dependencies, we will run this command:mix deps.getFor the second step, we need to define a Cache so that we can use it within the application's code. Let's define Blog.Cache module within lib/blog/cache.ex:defmodule Blog.Cache do use Nebulex.Cache, otp_app: :blog, adapter: Nebulex.Adapters.Local endThis module is what we'll be using to interact with the cache. It uses the Nebulex.Cache module and it expects the :otp_app as option. The otp_app tells Nebulex which Elixir application it can look for cache configuration in. In this case, we've specified that it is the :blog application where Nebulex can find that configuration and so Nebulex will use the configuration that was set up in config/config.exs.Could be configured in config/config.exs like so:config :blog, Blog.Cache, gc_interval: 86_400_000, #=&gt; 1 day backend: :shards, partitions: System.schedulers_online() #=&gt; The defaultFor more information about the provided options, see the adapter's documentation.The final piece of configuration is to setup the Blog.Cache as a supervisor within the application's supervision tree, which we can do in lib/blog/application.ex (or lib/blog.ex for elixir versions &lt; 1.4.0), inside the start/2 function:Elixir &lt; 1.5.0:def start(_type, _args) do import Supervisor.Spec, warn: false children = [ supervisor(Blog.Cache, []) ] ...Elixir &gt;= 1.5.0:def start(_type, _args) do children = [ Blog.Cache ] ...This piece of configuration will start the Nebulex process which receives and executes our application's commands. Without it, we wouldn't be able to use the cache at all!We've now configured our application so that it's able to execute commands against our cache.","ref":"getting-started.html#adding-nebulex-to-an-application","title":"Getting Started - Adding Nebulex to an application","type":"extras"},{"doc":"We can insert a new entries into our blog cache with this code:iex&gt; user = %{id: 1, first_name: &quot;Galileo&quot;, last_name: &quot;Galilei&quot;} iex&gt; Blog.Cache.put(user[:id], user, ttl: Nebulex.Time.expiry_time(1, :hour)) :okTo insert the data into our cache, we call put on Blog.Cache. This function tells Nebulex that we want to insert a new key/value entry into the cache corresponding Blog.Cache.It is also possible to insert multiple entries at once:iex&gt; users = %{ ...&gt; 1 =&gt; %{id: 1, first_name: &quot;Galileo&quot;, last_name: &quot;Galilei&quot;}, ...&gt; 2 =&gt; %{id: 2, first_name: &quot;Charles&quot;, last_name: &quot;Darwin&quot;}, ...&gt; 3 =&gt; %{id: 3, first_name: &quot;Albert&quot;, last_name: &quot;Einstein&quot;} ...&gt; } iex&gt; Blog.Cache.put_all(users) :okThe given entries can be a map or a Key/Value tuple list.Inserting new entries and replacing existing onesAs we saw previously, put creates a new entry in cache if it doesn't exist, or overrides it if it does exist (including the :ttl). However, there might be circumstances where we want to set the entry only if it doesn't exit or the other way around, this is where put_new and replace functions come in.Let's try put_new and put_new! functions:iex&gt; new_user = %{id: 4, first_name: &quot;John&quot;, last_name: &quot;Doe&quot;} iex&gt; Blog.Cache.put_new(new_user.id, new_user, ttl: 900) true iex&gt; Blog.Cache.put_new(new_user.id, new_user) false # same as previous one but raises `Nebulex.KeyAlreadyExistsError` iex&gt; Blog.Cache.put_new!(new_user.id, new_user)Now replace and replace! functions:iex&gt; existing_user = %{id: 5, first_name: &quot;John&quot;, last_name: &quot;Doe2&quot;} iex&gt; Blog.Cache.replace(existing_user.id, existing_user) false iex&gt; Blog.Cache.put_new(existing_user.id, existing_user) true iex&gt; Blog.Cache.replace(existing_user.id, existing_user, ttl: 900) true # same as previous one but raises `KeyError` iex&gt; Blog.Cache.replace!(100, existing_user)It is also possible to insert multiple new entries at once:iex&gt; new_users = %{ ...&gt; 6 =&gt; %{id: 6, first_name: &quot;Isaac&quot;, last_name: &quot;Newton&quot;}, ...&gt; 7 =&gt; %{id: 7, first_name: &quot;Marie&quot;, last_name: &quot;Curie&quot;} ...&gt; } iex&gt; Blog.Cache.put_new_all(new_users) true # none of the entries is inserted if at least one key already exists iex&gt; Blog.Cache.put_new_all(new_users) false","ref":"getting-started.html#inserting-entries","title":"Getting Started - Inserting entries","type":"extras"},{"doc":"Let’s start off with fetching data by the key, which is the most basic and common operation to retrieve data from a cache.iex&gt; Blog.Cache.get(1) _user_1 iex&gt; for key &lt;- 1..3 do ...&gt; user = Blog.Cache.get(key) ...&gt; user.first_name ...&gt; end [&quot;Galileo&quot;, &quot;Charles&quot;, &quot;Albert&quot;]There is a function has_key? to check if a key exist in cache:iex&gt; Blog.Cache.has_key?(1) true iex&gt; Blog.Cache.has_key?(10) falseRetrieving multiple entriesiex&gt; Blog.Cache.get_all([1, 2, 3]) _users","ref":"getting-started.html#retrieving-entries","title":"Getting Started - Retrieving entries","type":"extras"},{"doc":"Nebulex provides update and get_and_update functions to update an entry value based on current one, for example:iex&gt; initial = %{id: 1, first_name: &quot;&quot;, last_name: &quot;&quot;} # using `get_and_update` iex&gt; Blog.Cache.get_and_update(1, fn v -&gt; ...&gt; if v, do: {v, %{v | first_name: &quot;X&quot;}}, else: {v, initial} ...&gt; iex&gt; end) {_old, _updated} # using `update` iex&gt; Blog.Cache.update(1, initial, &amp;(%{&amp;1 | first_name: &quot;Y&quot;})) _updated","ref":"getting-started.html#updating-entries","title":"Getting Started - Updating entries","type":"extras"},{"doc":"The function incr is provided to increment or decrement a counter; by default, a counter is initialized to 0. Let's see how counters works:# by default, the counter is incremented by 1 iex&gt; Blog.Cache.incr(:my_counter) 1 # but we can also provide a custom increment value iex&gt; Blog.Cache.incr(:my_counter, 5) 6 # to decrement the counter, just pass a negative value iex&gt; Blog.Cache.incr(:my_counter, -5) 1","ref":"getting-started.html#counters","title":"Getting Started - Counters","type":"extras"},{"doc":"We’ve now covered inserting, reading and updating entries. Now let's see how to delete an entry using Nebulex.iex&gt; Blog.Cache.delete(1) :okTakeThis is another way not only for deleting an entry but also for retrieving it before its delete it:iex&gt; Blog.Cache.take(1) _entry # returns `nil` if `key` doesn&#39;t exist iex&gt; Blog.Cache.take(&quot;nonexistent&quot;) nil # same as previous one but raises `KeyError` iex&gt; Blog.Cache.take!(&quot;nonexistent&quot;)FlushNebulex also provides a function to flush all cache entries, like so:iex&gt; Blog.Cache.flush() :ok","ref":"getting-started.html#deleting-entries","title":"Getting Started - Deleting entries","type":"extras"},{"doc":"The last thing we’ll cover in this guide is how to retrieve information about cached objects or the cache itself.Remaining TTLiex&gt; Blog.Cache.ttl(1) _remaining_ttl iex&gt; Blog.Cache.ttl(&quot;nonexistent&quot;) nilCache sizeTo get the total number of cached objects:iex&gt; Blog.Cache.size() _num_cached_entries","ref":"getting-started.html#info","title":"Getting Started - Info","type":"extras"},{"doc":"Nebulex provides functions to fetch or stream all entries from cache matching the given query.To fetch all entries from cache:# by default, returns all keys iex&gt; Blog.Cache.all() _all_entries # fetch all entries and return the keys iex&gt; Blog.Cache.all(nil, return: :key) _keys # built-in queries in `Nebulex.Adapters.Local` adapter iex&gt; Blog.Cache.all(nil) iex&gt; Blog.Cache.all(:unexpired) iex&gt; Blog.Cache.all(:expired) # if we are using `Nebulex.Adapters.Local` adapter, the stored entry # is a tuple `{key, value, version, expire_at}`, then the match spec # could be something like: iex&gt; spec = [{{:&quot;$1&quot;, :&quot;$2&quot;, :_, :_}, [{:&gt;, :&quot;$2&quot;, 10}], [{{:&quot;$1&quot;, :&quot;$2&quot;}}]}] iex&gt; Blog.Cache.all(spec) _all_matched # using Ex2ms iex&gt; import Ex2ms iex&gt; spec = ...&gt; fun do ...&gt; {key, value, _, _} when value &gt; 10 -&gt; {key, value} ...&gt; end iex&gt; Blog.Cache.all(spec) _all_matchedIn the same way, we can stream all entries:iex&gt; Blog.Cache.stream() iex&gt; Blog.Cache.stream(nil, page_size: 3, return: :value) iex&gt; Blog.Cache.stream(nil, page_size: 3, return: :entry) # using `Nebulex.Adapters.Local` adapter iex&gt; spec = [{{:&quot;$1&quot;, :&quot;$2&quot;, :_, :_}, [{:&gt;, :&quot;$2&quot;, 10}], [{{:&quot;$1&quot;, :&quot;$2&quot;}}]}] iex&gt; Blog.Cache.stream(spec, page_size: 3) _all_matched # using Ex2ms iex&gt; import Ex2ms iex&gt; spec = ...&gt; fun do ...&gt; {key, value, _, _} when value &gt; 10 -&gt; {key, value} ...&gt; end iex&gt; Blog.Cache.stream(spec, page_size: 3) _all_matched","ref":"getting-started.html#query-and-or-stream-entires","title":"Getting Started - Query and/or Stream entires","type":"extras"},{"doc":"Nebulex provides the adapter Nebulex.Adapters.Partitioned, which allows to set up a partitioned cache topology.Let's define the Blog.PartitionedCache (lib/blog/partitioned_cache.ex) like so:defmodule Blog.PartitionedCache do use Nebulex.Cache, otp_app: :blog, adapter: Nebulex.Adapters.Partitioned endCould be configured with (config/config.exs):config :blog, Blog.PartitionedCache, primary: [ gc_interval: 86_400_000, backend: :shards, partitions: 2 ]And remember to setup the Blog.PartitionedCache as supervisor within the application's supervision tree (such as we did it previously):Elixir &lt; 1.5.0:def start(_type, _args) do import Supervisor.Spec, warn: false children = [ supervisor(Blog.PartitionedCache, []) ] ...Elixir &gt;= 1.5.0:def start(_type, _args) do import Supervisor.Spec, warn: false children = [ Blog.PartitionedCache ] ...Now we are ready to start using our partitioned cache!Timeout optionThe Nebulex.Adapters.Partitioned supports :timeout option, it is a value in milliseconds for the command that will be executed.iex&gt; Blog.PartitionedCache.get(&quot;foo&quot;, timeout: 10) _value # if the timeout is exceeded, then the current process will exit iex&gt; Blog.PartitionedCache.put(&quot;foo&quot;, &quot;bar&quot;, timeout: 10) # ** (EXIT) time outTo learn more about how partitioned cache works, please check Nebulex.Adapters.Partitioned documentation, and also it is recommended see the partitioned cache example","ref":"getting-started.html#partitioned-cache","title":"Getting Started - Partitioned Cache","type":"extras"},{"doc":"Nebulex also provides the adapter Nebulex.Adapters.Multilevel, which allows to setup a multi-level caching hierarchy.First, let's define out multi-level cache Blog.MultilevelCache:defmodule Blog.MultilevelCache do use Nebulex.Cache, otp_app: :blog, adapter: Nebulex.Adapters.Multilevel endCould be configured with (config/config.exs):config :blog, Blog.MultilevelCache, model: :inclusive, levels: [ l1: [ gc_interval: 86_400_000, backend: :shards ], l2: [ adapter: Nebulex.Adapters.Partitioned, primary: [ gc_interval: 86_400_000, backend: :shards ] ] ]And remember to setup the Blog.Multilevel as a supervisor within the application's supervision tree (such as we did it previously):Elixir &lt; 1.5.0:def start(_type, _args) do import Supervisor.Spec, warn: false children = [ supervisor(Blog.MultilevelCache, []) ] ...Elixir &gt;= 1.5.0:def start(_type, _args) do import Supervisor.Spec, warn: false children = [ Blog.Multilevel ] ...Let's try it out!iex&gt; Blog.Multilevel.put(&quot;foo&quot;, &quot;bar&quot;, ttl: Nebulex.Time.expiry_time(1, :hour)) &quot;bar&quot; iex&gt; Blog.Multilevel.get(&quot;foo&quot;) &quot;bar&quot;To learn more about how multilevel-cache works, please check Nebulex.Adapters.Multilevel documentation, and also it is recommended see the near cache example","ref":"getting-started.html#multilevel-cache","title":"Getting Started - Multilevel Cache","type":"extras"},{"doc":"Cache Usage Patterns via Nebulex.Caching - Annotations-based DSL to implement different cache usage patterns.","ref":"getting-started.html#next","title":"Getting Started - Next","type":"extras"},{"doc":"Cache Usage Patterns via Nebulex.CachingThere are several common access patterns when using a cache. Nebulex supports most of these patterns by means of Nebulex.Caching.Most of the following documentation about caching patterns it based on EHCache Docs","ref":"cache-usage-patterns.html","title":"Cache Usage Patterns via Nebulex.Caching","type":"extras"},{"doc":"With the cache-aside pattern, application code uses the cache directly.This means that application code which accesses the system-of-record (SoR) should consult the cache first, and if the cache contains the data, then return the data directly from the cache, bypassing the SoR. Otherwise, the application code must fetch the data from the system-of-record, store the data in the cache, and then return it. When data is written, the cache must be updated along with the system-of-record.Reading valuesif value = MyCache.get(key) do value else value = SoR.get(key) # maybe Ecto? :ok = MyCache.put(key, value) value endWriting values:ok = MyCache.put(key, value) SoR.insert(key, value) # maybe Ecto?As you may have noticed, this is the default behavior for most of the caches, we have to interact directly with the cache as well as the SoR (most likely the Database).","ref":"cache-usage-patterns.html#cache-aside","title":"Cache Usage Patterns via Nebulex.Caching - Cache-aside","type":"extras"},{"doc":"The cache-as-SoR pattern implies using the cache as though it were the primary system-of-record (SoR).The pattern delegates SoR reading and writing activities to the cache, so that application code is (at least directly) absolved of this responsibility. To implement the cache-as-SoR pattern, use a combination of the following read and write patterns:Read-throughWrite-throughAdvantages of using the cache-as-SoR pattern are:Less cluttered application code (improved maintainability through centralized SoR read/write operations)Choice of write-through or write-behind strategies on a per-cache basisAllows the cache to solve the thundering-herd problemA disadvantage of using the cache-as-SoR pattern is:Less directly visible code-pathBut how to get all this out-of-box? This is where the [Decorators][Decorators] comes in. It provides a set of macros to abstract most of the logic behind Read-through and Write-through patterns and make the implementation extremely easy. But let's go over these patterns more in detail and how to implement them using [Nebulex Decorators][Decorators].","ref":"cache-usage-patterns.html#cache-as-sor","title":"Cache Usage Patterns via Nebulex.Caching - Cache-as-SoR","type":"extras"},{"doc":"Under the read-through pattern, the cache is configured with a loader component that knows how to load data from the system-of-record (SoR).When the cache is asked for the value associated with a given key and such an entry does not exist within the cache, the cache invokes the loader to retrieve the value from the SoR, then caches the value, then returns it to the caller.The next time the cache is asked for the value for the same key it can be returned from the cache without using the loader (unless the entry has been evicted or expired).This pattern can be easily implemented using cache decorator as follows:defmodule MyApp.Example do use Nebulex.Caching alias MyApp.Cache @ttl Nebulex.Time.expiry_time(1, :hour) @decorate cacheable(cache: Cache, key: name) def get_by_name(name) do # your logic (the loader to retrieve the value from the SoR) end @decorate cacheable(cache: Cache, key: age, opts: [ttl: @ttl]) def get_by_age(age) do # your logic (the loader to retrieve the value from the SoR) end @decorate cacheable(cache: Cache) def all(query) do # your logic (the loader to retrieve the value from the SoR) end endAs you can see, the loader to retrieve the value from the system-of-record (SoR) is the function logic itself.","ref":"cache-usage-patterns.html#read-through","title":"Cache Usage Patterns via Nebulex.Caching - Read-through","type":"extras"},{"doc":"Under the write-through pattern, the cache is configured with a writer component that knows how to write data to the system-of-record (SoR).When the cache is asked to store a value for a key, the cache invokes the writer to store the value in the SoR, as well as updating (or deleting) the cache.This pattern can be implemented using defevict or defupdatable. When the data is written to the system-of-record (SoR), you can update the cached value associated with the given key using defupdatable, or just delete it using defevict.defmodule MyApp.Example do use Nebulex.Caching alias MyApp.Cache # When the data is written to the SoR, it is updated in the cache @decorate cache_put(cache: Cache, key: something) def update(something) do # Write data to the SoR (most likely the Database) end # When the data is written to the SoR, it is deleted (evicted) from the cache @decorate cache_evict(cache: Cache, key: something) def update_something(something) do # Write data to the SoR (most likely the Database) end endAs you can see, the logic to write data to the system-of-record (SoR) is the function logic itself.","ref":"cache-usage-patterns.html#write-through","title":"Cache Usage Patterns via Nebulex.Caching - Write-through","type":"extras"},{"doc":"TelemetryThis guide aims to show you how to instrument and report on :telemetry events in your application for cache statistics.This guide is not focused on explaining :telemetry, how it works, configure it, and so on. Instead, it is focused on how we can use :telemetry for reporting cache stats. For more information about :telemetry you can check the docs, or the Phoenix Telemetry guide is also recommended.","ref":"telemetry.html","title":"Telemetry","type":"extras"},{"doc":"The stats support is something each adapter is responsible for. However, Nebulex built-in adapters support the stats suggested and defined by Nebulex.Cache.Stats. Besides, when the :stats option is enabled, we can use Telemetry for emitting the current stat values.First of all, let's configure the dependencies adding :telemetry_metrics and :telemetry_poller packages:def deps do [ {:nebulex, &quot;~&gt; 2.0&quot;}, {:shards, &quot;~&gt; 0.6&quot;}, #=&gt; For using :shards as backend {:decorator, &quot;~&gt; 1.3&quot;} #=&gt; For using Caching Annotations {:telemetry_metrics, &quot;~&gt; 0.5&quot;}, {:telemetry_poller, &quot;~&gt; 0.5&quot;} ] endThen define the cache and add the configuration:defmodule MyApp.Cache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local # Use stats helpers use Nebulex.Cache.Stats endCould be configured:config :my_app, MyApp.Cache, stats: true, backend: :shards, gc_interval: 86_400_000, max_size: 1_000_000, gc_cleanup_min_timeout: 10_000, gc_cleanup_max_timeout: 900_000Create your Telemetry supervisor at lib/my_app/telemetry.ex:# lib/my_app/telemetry.ex defmodule MyApp.Telemetry do use Supervisor import Telemetry.Metrics def start_link(arg) do Supervisor.start_link(__MODULE__, arg, name: __MODULE__) end def init(_arg) do children = [ # Configure `:telemetry_poller` for reporting the cache stats {:telemetry_poller, measurements: periodic_measurements(), period: 10_000} # For example, we use the console reporter, but you can change it. # See `:telemetry_metrics` for for information. {Telemetry.Metrics.ConsoleReporter, metrics: metrics()} ] Supervisor.init(children, strategy: :one_for_one) end defp metrics do [ # Nebulex Stats Metrics last_value(&quot;nebulex.cache.stats.hits&quot;, tags: [:cache]), last_value(&quot;nebulex.cache.stats.misses&quot;, tags: [:cache]), last_value(&quot;nebulex.cache.stats.writes&quot;, tags: [:cache]), last_value(&quot;nebulex.cache.stats.evictions&quot;, tags: [:cache]), last_value(&quot;nebulex.cache.stats.expirations&quot;, tags: [:cache]), # VM Metrics summary(&quot;vm.memory.total&quot;, unit: {:byte, :kilobyte}), summary(&quot;vm.total_run_queue_lengths.total&quot;), summary(&quot;vm.total_run_queue_lengths.cpu&quot;), summary(&quot;vm.total_run_queue_lengths.io&quot;) ] end defp periodic_measurements do [ {MyApp.Cache, :dispatch_stats, []} ] end endMake sure to replace MyApp by your actual application name.Then add to your main application's supervision tree (usually in lib/my_app/application.ex):children = [ MyApp.Cache, MyAppWeb.Telemetry, ... ]Now start an IEx session and call the server:iex(1)&gt; MyApp.Cache.get 1 nil iex(2)&gt; MyApp.Cache.put 1, 1, ttl: 10 :ok iex(3)&gt; MyApp.Cache.get 1 1 iex(4)&gt; MyApp.Cache.put 2, 2 :ok iex(5)&gt; MyApp.Cache.delete 2 :ok iex(6)&gt; Process.sleep(20) :ok iex(7)&gt; MyApp.Cache.get 1 niland you should see something like the following output:[Telemetry.Metrics.ConsoleReporter] Got new event! Event name: nebulex.cache.stats All measurements: %{evictions: 2, expirations: 1, hits: 1, misses: 2, writes: 2} All metadata: %{cache: MyApp.Cache} Metric measurement: :hits (last_value) With value: 1 Tag values: %{cache: MyApp.Cache} Metric measurement: :misses (last_value) With value: 2 Tag values: %{cache: MyApp.Cache} Metric measurement: :writes (last_value) With value: 2 Tag values: %{cache: MyApp.Cache} Metric measurement: :evictions (last_value) With value: 2 Tag values: %{cache: MyApp.Cache} Metric measurement: :expirations (last_value) With value: 1 Tag values: %{cache: MyApp.Cache}","ref":"telemetry.html#instrumenting-nebulex-caches","title":"Telemetry - Instrumenting Nebulex Caches","type":"extras"},{"doc":"In the same way, you can, for instance, add another periodic measurement for reporting the cache size.Using our previous cache:defmodule MyApp.Cache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local # Use stats helpers use Nebulex.Cache.Stats def dispatch_cache_size do :telemetry.execute( [:nebulex, :cache, :size], %{value: size()}, %{cache: __MODULE__} ) end endAnd add it to the list of periodic measurements in our previously defined supervisor:defp periodic_measurements do [ {MyApp.Cache, :dispatch_stats, []}, {MyApp.Cache, :dispatch_cache_size, []} ] endMetrics:defp metrics do [ # Nebulex Stats Metrics last_value(&quot;nebulex.cache.stats.hits&quot;, tags: [:cache]), last_value(&quot;nebulex.cache.stats.misses&quot;, tags: [:cache]), last_value(&quot;nebulex.cache.stats.writes&quot;, tags: [:cache]), last_value(&quot;nebulex.cache.stats.evictions&quot;, tags: [:cache]), last_value(&quot;nebulex.cache.stats.expirations&quot;, tags: [:cache]), # Nebulex custom Metrics last_value(&quot;nebulex.cache.size.value&quot;, tags: [:cache]), # VM Metrics summary(&quot;vm.memory.total&quot;, unit: {:byte, :kilobyte}), summary(&quot;vm.total_run_queue_lengths.total&quot;), summary(&quot;vm.total_run_queue_lengths.cpu&quot;), summary(&quot;vm.total_run_queue_lengths.io&quot;) ] endIf you start an IEx session like previously, you should see the new metric too:[Telemetry.Metrics.ConsoleReporter] Got new event! Event name: nebulex.cache.size All measurements: %{value: 0} All metadata: %{cache: MyApp.Cache} Metric measurement: :value (last_value) With value: 0 Tag values: %{cache: MyApp.Cache}","ref":"telemetry.html#adding-other-custom-metrics","title":"Telemetry - Adding other custom metrics","type":"extras"},{"doc":"Migrating to v2.xFor the v2, Nebulex introduces several breaking changes, so this guide aims to highlight most of these changes to make easier the transition to v2. Be aware this guide won't focus on every change, just the most significant ones that can affect how your application code interacts with the cache. Also, it is not a detailed guide about how to translate the current code from older versions to v2, just pointing out the areas the new documentation should be consulted on.","ref":"migrating-to-v2.html","title":"Migrating to v2.x","type":"extras"},{"doc":"This is one of the biggest changes. Version 1.x, most of the configuration options are resolved in compile-time, but this brings with a lot of limitations. Since version 2.x, the only argument or option configured in compile-time is the adapter when defining a new cache module, the rest of them are given via config file or at startup time. For more information and examples, see Nebulex.Cache, Nebulex.Adapters.Local, Nebulex.Adapters.Partitioned, Nebulex.Adapters.Replicated, Nebulex.Adapters.Multilevel.","ref":"migrating-to-v2.html#configiration","title":"Migrating to v2.x - Configiration","type":"extras"},{"doc":"There are several changes on the Nebulex.Cache API:The :return option is not available anymore, so it has to be removed.The :version option is not available anymore, so it has to be removed.Callback c:set/3 was refactored to c:put/3.Callback c:set_many/2 was refactored to c:put_all/2.Callbacks c:add/3 and c:add!/3 were refactored to c:put_new/3 and c:put_new!/3.Callback c:update_counter/3 was refactored to c:incr/3.Callback c:add_or_replace/3 was removed.Callback c:object_info/2 was removed, and callbacks c:ttl/1 and c:touch/1 were added instead.","ref":"migrating-to-v2.html#cache-api","title":"Migrating to v2.x - Cache API","type":"extras"},{"doc":"Module Nebulex.Caching.Decorators was refactored to Nebulex.Caching – Keep in mind that since v1.2.x the caching decorators were included instead of the previous macros or DSL (this applies for version lower than v1.1.x).Decorator cache/3 was refactored to cacheable/3.Decorator evict/3 was refactored to cache_evict/3.Decorator update/3 was refactored to cache_put/3.Improved the :match option to return not only a boolean but return a specific value to be cached (term -&gt; boolean | {true, term}) – If true the code-block evaluation result is cached as it is (the default). If {true, value} is returned, then the value is what is cached.","ref":"migrating-to-v2.html#declarative-annotation-based-caching-via-decorators","title":"Migrating to v2.x - Declarative annotation-based caching via decorators","type":"extras"},{"doc":"Since v2.x, pre/post hooks are deprecated and won't be longer supported by Nebulex, at least not directly. Mainly, because the hooks feature is not a common use-case and also it is something that can be be easily implemented on top of the Cache at the application level. However, to keep backward compatibility somehow, Nebulex provides decorators for implementing pre/post hooks very easily. For that reason, it is highly recommended to removed all pre/post hooks related code and adapt it to the new way. See Nebulex.Hook for more information.","ref":"migrating-to-v2.html#hooks","title":"Migrating to v2.x - Hooks","type":"extras"},{"doc":"There have been several and significant improvements on the built-in adapters, so it is also highly recommended to take a look at them; Nebulex.Adapters.Local, Nebulex.Adapters.Partitioned, Nebulex.Adapters.Replicated, and Nebulex.Adapters.Multilevel.","ref":"migrating-to-v2.html#built-in-adapters","title":"Migrating to v2.x - Built-In Adapters","type":"extras"},{"doc":"For older versions (&lt;= 1.x), the stats were implemented via a post-hook and the measurements were oriented for counting the number of times a cache function is called. But what is interesting and useful to see is, for instance, the number of writes, hits, misses, evictions, etc. Therefore, the whole stats' functionality was refactored entirely.This feature is not longer using pre/post hooks. Besides, pre/post hooks are deprecated in v2.x.The stats support is something each adapter is responsible for. However, Nebulex built-in adapters support the stats suggested and defined by Nebulex.Cache.Stats.Default stats via Nebulex.Cache.Stats are implemented by means of Erlang :counters.See Nebulex.Cache.Stats for more information about how to use stats in Nebulex v2.x.","ref":"migrating-to-v2.html#statistics","title":"Migrating to v2.x - Statistics","type":"extras"}]